# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""Unit tests for :func:`iris.analysis.cartography.project`."""

# Import iris.tests first so that some things can be initialised before
# importing anything else.
import iris.tests as tests

import cartopy.crs as ccrs
import numpy as np

import iris.coord_systems
import iris.coords
import iris.cube
import iris.tests
import iris.tests.stock

from iris.analysis.cartography import project


ROBINSON = ccrs.Robinson()


def low_res_4d():
    cube = iris.tests.stock.realistic_4d_no_derived()
    cube = cube[0:2, 0:3, ::10, ::10]
    cube.remove_coord("surface_altitude")
    return cube


class TestAll(tests.IrisTest):
    def setUp(self):
        cs = iris.coord_systems.GeogCS(654321)
        self.cube = iris.cube.Cube(np.zeros(25).reshape(5, 5))
        self.cube.add_dim_coord(
            iris.coords.DimCoord(
                np.arange(5),
                standard_name="latitude",
                units="degrees",
                coord_system=cs,
            ),
            0,
        )
        self.cube.add_dim_coord(
            iris.coords.DimCoord(
                np.arange(5),
                standard_name="longitude",
                units="degrees",
                coord_system=cs,
            ),
            1,
        )

        self.tcs = iris.coord_systems.GeogCS(600000)

    def test_is_iris_coord_system(self):
        res, _ = project(self.cube, self.tcs)
        self.assertEqual(
            res.coord("projection_y_coordinate").coord_system, self.tcs
        )
        self.assertEqual(
            res.coord("projection_x_coordinate").coord_system, self.tcs
        )

        self.assertIsNot(
            res.coord("projection_y_coordinate").coord_system, self.tcs
        )
        self.assertIsNot(
            res.coord("projection_x_coordinate").coord_system, self.tcs
        )

    @tests.skip_data
    def test_bad_resolution_negative(self):
        cube = low_res_4d()
        with self.assertRaises(ValueError):
            project(cube, ROBINSON, nx=-200, ny=200)

    @tests.skip_data
    def test_bad_resolution_non_numeric(self):
        cube = low_res_4d()
        with self.assertRaises(TypeError):
            project(cube, ROBINSON, nx=200, ny="abc")

    @tests.skip_data
    def test_missing_lat(self):
        cube = low_res_4d()
        cube.remove_coord("grid_latitude")
        with self.assertRaises(ValueError):
            project(cube, ROBINSON)

    @tests.skip_data
    def test_missing_lon(self):
        cube = low_res_4d()
        cube.remove_coord("grid_longitude")
        with self.assertRaises(ValueError):
            project(cube, ROBINSON)

    @tests.skip_data
    def test_missing_latlon(self):
        cube = low_res_4d()
        cube.remove_coord("grid_longitude")
        cube.remove_coord("grid_latitude")
        with self.assertRaises(ValueError):
            project(cube, ROBINSON)

    @tests.skip_data
    def test_default_resolution(self):
        cube = low_res_4d()
        new_cube, extent = project(cube, ROBINSON)
        self.assertEqual(new_cube.shape, cube.shape)

    @tests.skip_data
    def test_explicit_resolution(self):
        cube = low_res_4d()
        nx, ny = 5, 4
        new_cube, extent = project(cube, ROBINSON, nx=nx, ny=ny)
        self.assertEqual(new_cube.shape, cube.shape[:2] + (ny, nx))

    @tests.skip_data
    def test_explicit_resolution_single_point(self):
        cube = low_res_4d()
        nx, ny = 1, 1
        new_cube, extent = project(cube, ROBINSON, nx=nx, ny=ny)
        self.assertEqual(new_cube.shape, cube.shape[:2] + (ny, nx))

    @tests.skip_data
    def test_mismatched_coord_systems(self):
        cube = low_res_4d()
        cube.coord("grid_longitude").coord_system = None
        with self.assertRaises(ValueError):
            project(cube, ROBINSON)

    @tests.skip_data
    def test_extent(self):
        cube = low_res_4d()
        _, extent = project(cube, ROBINSON)
        self.assertEqual(
            extent,
            [
                -17005833.33052523,
                17005833.33052523,
                -8625154.6651,
                8625154.6651,
            ],
        )

    @tests.skip_data
    def test_cube(self):
        cube = low_res_4d()
        new_cube, _ = project(cube, ROBINSON)
        self.assertCMLApproxData(new_cube)

    @tests.skip_data
    def test_no_coord_system(self):
        cube = low_res_4d()
        cube.coord("grid_longitude").coord_system = None
        cube.coord("grid_latitude").coord_system = None
        with iris.tests.mock.patch("warnings.warn") as warn:
            _, _ = project(cube, ROBINSON)
        warn.assert_called_once_with(
            "Coordinate system of latitude and "
            "longitude coordinates is not specified. "
            "Assuming WGS84 Geodetic."
        )


if __name__ == "__main__":
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.

import os
import sys
import re
import inspect
import ntpath

# list of tuples for modules to exclude.  Useful if the documentation throws
# warnings, especially for experimental modules.
exclude_modules = [
    ("experimental/raster", "iris.experimental.raster")  # gdal conflicts
]


# print to stdout, including the name of the python file
def autolog(message):
    print("[{}] {}".format(ntpath.basename(__file__), message))


document_dict = {
    # Use autoclass for classes.
    "class": """
{object_docstring}

..

    .. autoclass:: {object_name}
        :members:
        :undoc-members:
        :inherited-members:

""",
    "function": """
.. autofunction:: {object_name}

""",
    # For everything else, let automodule do some magic...
    None: """

.. autodata:: {object_name}

""",
}


horizontal_sep = """
.. raw:: html

    <p class="hr_p"><a href="#">&uarr;&#32&#32 top &#32&#32&uarr;</a></p>
    <!--

-----------

.. raw:: html

    -->

"""


def lookup_object_type(obj):
    if inspect.isclass(obj):
        return "class"
    elif inspect.isfunction(obj):
        return "function"
    else:
        return None


def auto_doc_module(
    file_path, import_name, root_package, package_toc=None, title=None
):
    doc = r""".. _{import_name}:

{title_underline}
{title}
{title_underline}

{sidebar}

.. currentmodule:: {root_package}

.. automodule:: {import_name}

In this module:

{module_elements}

"""

    if package_toc:
        sidebar = """
{package_toc_tree}

    """.format(
            package_toc_tree=package_toc
        )
    else:
        sidebar = ""

    try:
        mod = __import__(import_name)
    except ImportError as e:
        message = r""".. error::

    This module could not be imported. Some dependencies are missing::

        """ + str(
            e
        )
        return doc.format(
            title=title or import_name,
            title_underline="=" * len(title or import_name),
            import_name=import_name,
            root_package=root_package,
            sidebar=sidebar,
            module_elements=message,
        )

    mod = sys.modules[import_name]
    elems = dir(mod)

    if "__all__" in elems:
        document_these = [
            (attr_name, getattr(mod, attr_name)) for attr_name in mod.__all__
        ]
    else:
        document_these = [
            (attr_name, getattr(mod, attr_name))
            for attr_name in elems
            if (
                not attr_name.startswith("_")
                and not inspect.ismodule(getattr(mod, attr_name))
            )
        ]

        def is_from_this_module(arg):
            # name = arg[0]
            obj = arg[1]
            return (
                hasattr(obj, "__module__") and obj.__module__ == mod.__name__
            )

        sort_order = {"class": 2, "function": 1}

        # Sort them according to sort_order dict.
        def sort_key(arg):
            # name = arg[0]
            obj = arg[1]
            return sort_order.get(lookup_object_type(obj), 0)

        document_these = filter(is_from_this_module, document_these)
        document_these = sorted(document_these, key=sort_key)

    lines = []
    for element, obj in document_these:
        object_name = import_name + "." + element
        obj_content = document_dict[lookup_object_type(obj)].format(
            object_name=object_name,
            object_name_header_line="+" * len(object_name),
            object_docstring=inspect.getdoc(obj),
        )
        lines.append(obj_content)

    lines = horizontal_sep.join(lines)

    module_elements = "\n".join(
        " * :py:obj:`{}`".format(element) for element, obj in document_these
    )

    lines = doc + lines
    return lines.format(
        title=title or import_name,
        title_underline="=" * len(title or import_name),
        import_name=import_name,
        root_package=root_package,
        sidebar=sidebar,
        module_elements=module_elements,
    )


def auto_doc_package(file_path, import_name, root_package, sub_packages):
    max_depth = 1 if import_name == "iris" else 2
    package_toc = "\n      ".join(sub_packages)

    package_toc = """
   .. toctree::
      :maxdepth: {:d}
      :titlesonly:
      :hidden:

      {}


""".format(
        max_depth, package_toc
    )

    if "." in import_name:
        title = None
    else:
        title = import_name.capitalize() + " API"

    return auto_doc_module(
        file_path,
        import_name,
        root_package,
        package_toc=package_toc,
        title=title,
    )


def auto_package_build(app):
    root_package = app.config.autopackage_name
    if root_package is None:
        raise ValueError(
            "set the autopackage_name variable in the " "conf.py file"
        )

    if not isinstance(root_package, list):
        raise ValueError(
            "autopackage was expecting a list of packages to "
            'document e.g. ["itertools"]'
        )

    for package in root_package:
        do_package(package)


def do_package(package_name):
    out_dir = "generated/api" + os.path.sep

    # Import the root package. If this fails then an import error will be
    # raised.
    module = __import__(package_name)
    root_package = package_name
    rootdir = os.path.dirname(module.__file__)

    package_folder = []
    module_folders = {}

    for root, subFolders, files in os.walk(rootdir):
        for fname in files:
            name, ext = os.path.splitext(fname)

            # Skip some non-relevant files.
            if (
                fname.startswith(".")
                or fname.startswith("#")
                or re.search("^_[^_]", fname)
                or fname.find(".svn") >= 0
                or not (ext in [".py", ".so"])
            ):
                continue

            # Handle new shared library naming conventions
            if ext == ".so":
                name = name.split(".", 1)[0]

            rel_path = (
                root_package + os.path.join(root, fname).split(rootdir)[-1]
            )
            mod_folder = root_package + os.path.join(root).split(rootdir)[
                -1
            ].replace("/", ".")

            # Only add this package to folder list if it contains an __init__
            # script.
            if name == "__init__":
                package_folder.append([mod_folder, rel_path])
            else:
                import_name = mod_folder + "." + name
                mf_list = module_folders.setdefault(mod_folder, [])
                mf_list.append((import_name, rel_path))
    if not os.path.exists(out_dir):
        os.makedirs(out_dir)

    for package, package_path in package_folder:
        if "._" in package or "test" in package:
            continue

        paths = []
        for spackage, spackage_path in package_folder:
            # Ignore this packages, packages that are not children of this
            # one, test packages, private packages, and packages that are
            # subpackages of subpackages (they'll be part of the subpackage).
            if spackage == package:
                continue
            if not spackage.startswith(package):
                continue
            if spackage.count(".") > package.count(".") + 1:
                continue
            if "test" in spackage:
                continue

            split_path = spackage.rsplit(".", 2)[-2:]
            if any(part[0] == "_" for part in split_path):
                continue

            paths.append(os.path.join(*split_path) + ".rst")

        paths.extend(
            os.path.join(
                os.path.basename(os.path.dirname(path)),
                os.path.basename(path).split(".", 1)[0],
            )
            for imp_name, path in module_folders.get(package, [])
        )

        paths.sort()

        # check for any modules to exclude
        for exclude_module in exclude_modules:
            if exclude_module[0] in paths:
                autolog(
                    "Excluding module in package: {}".format(exclude_module[0])
                )
                paths.remove(exclude_module[0])

        doc = auto_doc_package(package_path, package, root_package, paths)

        package_dir = out_dir + package.replace(".", os.path.sep)
        if not os.path.exists(package_dir):
            os.makedirs(out_dir + package.replace(".", os.path.sep))

        out_path = package_dir + ".rst"
        if not os.path.exists(out_path):
            autolog("Creating {} ...".format(out_path))
            with open(out_path, "w") as fh:
                fh.write(doc)
        else:
            with open(out_path, "r") as fh:
                existing_content = "".join(fh.readlines())
            if doc != existing_content:
                autolog("Creating {} ...".format(out_path))
                with open(out_path, "w") as fh:
                    fh.write(doc)

        for import_name, module_path in module_folders.get(package, []):
            # check for any modules to exclude
            for exclude_module in exclude_modules:
                if import_name == exclude_module[1]:
                    autolog(
                        "Excluding module file: {}".format(exclude_module[1])
                    )
                else:
                    doc = auto_doc_module(
                        module_path, import_name, root_package
                    )
                    out_path = (
                        out_dir
                        + import_name.replace(".", os.path.sep)
                        + ".rst"
                    )
                    if not os.path.exists(out_path):
                        autolog("Creating {} ...".format(out_path))
                        with open(out_path, "w") as fh:
                            fh.write(doc)
                    else:
                        with open(out_path, "r") as fh:
                            existing_content = "".join(fh.readlines())
                        if doc != existing_content:
                            autolog("Creating {} ...".format(out_path))
                            with open(out_path, "w") as fh:
                                fh.write(doc)


def setup(app):
    app.connect("builder-inited", auto_package_build)
    app.add_config_value("autopackage_name", None, "env")



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Unit tests for the
:mod:`iris.fileformats._structured_array_identification.ArrayStructure` class.

"""

# Import iris.tests first so that some things can be initialised before
# importing anything else.
import iris.tests as tests

import numpy as np

from iris.fileformats._structured_array_identification import (
    ArrayStructure,
    _UnstructuredArrayException,
)


def construct_nd(sub_array, sub_dim, shape):
    # Given a 1D array, a shape, and the axis/dimension that the 1D array
    # represents on the bigger array, construct a numpy array which is
    # filled appropriately.
    assert sub_array.ndim == 1
    sub_shape = [1 if dim != sub_dim else -1 for dim in range(len(shape))]
    return sub_array.reshape(sub_shape) * np.ones(shape)


class TestArrayStructure_from_array(tests.IrisTest):
    def struct_from_arr(self, nd_array):
        return ArrayStructure.from_array(nd_array.flatten())

    def test_1d_len_0(self):
        a = np.arange(0)
        self.assertEqual(self.struct_from_arr(a), ArrayStructure(1, a))

    def test_1d_len_1(self):
        a = np.arange(1)
        self.assertEqual(self.struct_from_arr(a), ArrayStructure(1, a))

    def test_1d(self):
        a = np.array([-1, 3, 1, 2])
        self.assertEqual(self.struct_from_arr(a), ArrayStructure(1, a))

    def test_1d_ones(self):
        a = np.ones(10)
        self.assertEqual(self.struct_from_arr(a), ArrayStructure(1, [1]))

    def test_1d_range(self):
        a = np.arange(6)
        self.assertEqual(
            self.struct_from_arr(a), ArrayStructure(1, list(range(6)))
        )

    def test_3d_ones(self):
        a = np.ones([10, 2, 1])
        self.assertEqual(self.struct_from_arr(a), ArrayStructure(1, [1]))

    def test_1d_over_2d_first_dim_manual(self):
        sub = np.array([10, 10, 20, 20])
        self.assertEqual(
            self.struct_from_arr(sub), ArrayStructure(2, [10, 20])
        )

    def test_3d_first_dimension(self):
        flattened = np.array([1, 1, 1, 2, 2, 2])
        self.assertEqual(
            ArrayStructure.from_array(flattened), ArrayStructure(3, [1, 2])
        )

    def test_1d_over_2d_first_dim(self):
        sub = np.array([-1, 3, 1, 2])
        a = construct_nd(sub, 0, (4, 2))
        self.assertEqual(self.struct_from_arr(a), ArrayStructure(2, sub))

    def test_1d_over_2d_second_dim(self):
        sub = np.array([-1, 3, 1, 2])
        a = construct_nd(sub, 1, (2, 4))
        self.assertEqual(self.struct_from_arr(a), ArrayStructure(1, sub))

    def test_1d_over_3d_first_dim(self):
        sub = np.array([-1, 3, 1, 2])
        a = construct_nd(sub, 0, (4, 2, 3))
        self.assertEqual(self.struct_from_arr(a), ArrayStructure(6, sub))

    def test_1d_over_3d_second_dim(self):
        sub = np.array([-1, 3, 1, 2])
        a = construct_nd(sub, 1, (2, 4, 3))
        self.assertEqual(self.struct_from_arr(a), ArrayStructure(3, sub))

    def test_1d_over_3d_third_dim(self):
        sub = np.array([-1, 3, 1, 2])
        a = construct_nd(sub, 2, (3, 2, 4))
        self.assertEqual(self.struct_from_arr(a), ArrayStructure(1, sub))

    def test_irregular_3d(self):
        sub = np.array([-1, 3, 1, 2])
        a = construct_nd(sub, 2, (3, 2, 4))
        a[0, 0, 0] = 5
        self.assertEqual(self.struct_from_arr(a), None)

    def test_repeated_3d(self):
        sub = np.array([-1, 3, 1, 2])
        a = construct_nd(sub, 2, (3, 2, 4))
        a[:, 0, 0] = 1
        self.assertEqual(self.struct_from_arr(a), None)

    def test_rolled_3d(self):
        # Shift the 3D array on by one, making the array 1d.
        sub = np.arange(4)
        a = construct_nd(sub, 0, (4, 2, 3))
        a = np.roll(a.flatten(), 1)
        self.assertEqual(self.struct_from_arr(a), None)

    def test_len_1_3d(self):
        # Setup a case which triggers an IndexError when identifying
        # the stride, but the result should still be correct.
        sub = np.arange(2)
        a = construct_nd(sub, 1, (1, 1, 1))
        self.assertEqual(self.struct_from_arr(a), ArrayStructure(1, sub))

    def test_not_an_array(self):
        # Support lists as an argument.
        self.assertEqual(
            ArrayStructure.from_array([1, 2, 3]), ArrayStructure(1, [1, 2, 3])
        )

    def test_multi_dim_array(self):
        with self.assertRaises(ValueError):
            ArrayStructure.from_array(np.arange(12).reshape(3, 4))


class nd_array_and_dims_cases:
    """
    Defines the test functionality for nd_array_and_dims. This class
    isn't actually the test case - see the C order and F order subclasses
    for those.

    """

    def test_scalar_len1_first_dim(self):
        struct = ArrayStructure(1, [1])
        orig = np.array([1, 1, 1])

        array, dims = struct.nd_array_and_dims(orig, (1, 3), order=self.order)
        self.assertArrayEqual(array, [1])
        self.assertEqual(dims, ())

    def test_scalar_non_len1_first_dim(self):
        struct = ArrayStructure(1, [1])
        orig = np.array([1, 1, 1])

        array, dims = struct.nd_array_and_dims(orig, (3, 1), order=self.order)
        self.assertArrayEqual(array, [1])
        self.assertEqual(dims, ())

    def test_single_vector(self):
        orig = construct_nd(np.array([1, 2]), 0, (2, 1, 3))
        flattened = orig.flatten(order=self.order)
        struct = ArrayStructure.from_array(flattened)
        array, dims = struct.nd_array_and_dims(
            flattened, (2, 1, 3), order=self.order
        )
        self.assertArrayEqual(array, [1, 2])
        self.assertEqual(dims, (0,))

    def test_single_vector_3rd_dim(self):
        orig = construct_nd(np.array([1, 2, 3]), 2, (4, 1, 3))
        flattened = orig.flatten(order=self.order)

        struct = ArrayStructure.from_array(flattened)
        array, dims = struct.nd_array_and_dims(
            flattened, (4, 1, 3), order=self.order
        )
        self.assertArrayEqual(array, [1, 2, 3])
        self.assertEqual(dims, (2,))

    def test_orig_array_and_target_shape_inconsistent(self):
        # An array structure which has a length which is a product
        # of potential dimensions should not result in an array
        struct = ArrayStructure(2, [1, 2, 3])
        orig = np.array([1, 1, 2, 2, 3, 3])

        msg = "Original array and target shape do not match up."
        with self.assertRaisesRegex(ValueError, msg):
            struct.nd_array_and_dims(orig, (2, 3, 2), order=self.order)

    def test_array_bigger_than_expected(self):
        # An array structure which has a length which is a product
        # of potential dimensions should not result in an array
        struct = ArrayStructure(2, [1, 2, 3, 4, 5, 6])
        orig = np.array([1, 1, 2, 2, 3, 3, 4, 4, 5, 5, 6, 6])

        with self.assertRaises(_UnstructuredArrayException):
            struct.nd_array_and_dims(orig, (2, 3, 2), order=self.order)

    def test_single_vector_extra_dimension(self):
        orig = construct_nd(np.array([1, 2]), 1, (3, 2))
        flattened = orig.flatten(order=self.order)

        struct = ArrayStructure.from_array(flattened)

        # Add another dimension on flattened, making it a (6, 2).
        input_array = np.vstack([flattened, flattened + 100]).T

        array, dims = struct.nd_array_and_dims(
            input_array, (3, 1, 2, 1), order=self.order
        )
        self.assertArrayEqual(array, [[1, 101], [2, 102]])
        self.assertEqual(dims, (2,))


class TestArrayStructure_nd_array_and_dims_f_order(
    tests.IrisTest, nd_array_and_dims_cases
):
    order = "f"


class TestArrayStructure_nd_array_and_dims_c_order(
    tests.IrisTest, nd_array_and_dims_cases
):
    order = "c"


if __name__ == "__main__":
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.


# Historically this was auto-generated from
# SciTools/iris-code-generators:tools/gen_rules.py

import cf_units
import numpy as np
import calendar

from iris.aux_factory import HybridHeightFactory, HybridPressureFactory
from iris.coords import AuxCoord, CellMethod, DimCoord
from iris.fileformats.rules import (
    ConversionMetadata,
    Factory,
    Reference,
    ReferenceTarget,
)
from iris.fileformats._pp_lbproc_pairs import LBPROC_MAP
from iris.fileformats.um_cf_map import (
    LBFC_TO_CF,
    STASH_TO_CF,
    STASHCODE_IMPLIED_HEIGHTS,
)


###############################################################################
#
# Convert vectorisation routines.
#


def _dim_or_aux(*args, **kwargs):
    try:
        result = DimCoord(*args, **kwargs)
    except ValueError:
        attr = kwargs.get("attributes")
        if attr is not None and "positive" in attr:
            del attr["positive"]
        result = AuxCoord(*args, **kwargs)
    return result


def _convert_vertical_coords(
    lbcode,
    lbvc,
    blev,
    lblev,
    stash,
    bhlev,
    bhrlev,
    brsvd1,
    brsvd2,
    brlev,
    dim=None,
):
    """
    Encode scalar or vector vertical level values from PP headers as CM data
    components.

    Args:

    * lbcode:
        Scalar field :class:`iris.fileformats.pp.SplittableInt` value.

    * lbvc:
        Scalar field value.

    * blev:
        Scalar field value or :class:`numpy.ndarray` vector of field values.

    * lblev:
        Scalar field value or :class:`numpy.ndarray` vector of field values.

    * stash:
        Scalar field :class:`iris.fileformats.pp.STASH` value.

    * bhlev:
        Scalar field value or :class:`numpy.ndarray` vector of field values.

    * bhrlev:
        Scalar field value or :class:`numpy.ndarray` vector of field values.

    * brsvd1:
        Scalar field value or :class:`numpy.ndarray` vector of field values.

    * brsvd2:
        Scalar field value or :class:`numpy.ndarray` vector of field values.

    * brlev:
        Scalar field value or :class:`numpy.ndarray` vector of field values.

    Kwargs:

    * dim:
        Associated dimension of the vertical coordinate. Defaults to None.

    Returns:
        A tuple containing a list of coords_and_dims, and a list of factories.

    """
    factories = []
    coords_and_dims = []

    # See Word no. 33 (LBLEV) in section 4 of UM Model Docs (F3).
    BASE_RHO_LEVEL_LBLEV = 9999
    model_level_number = np.atleast_1d(lblev)
    model_level_number[model_level_number == BASE_RHO_LEVEL_LBLEV] = 0

    # Ensure to vectorise these arguments as arrays, as they participate
    # in the conditions of convert rules.
    blev = np.atleast_1d(blev)
    brsvd1 = np.atleast_1d(brsvd1)
    brlev = np.atleast_1d(brlev)

    # Height.
    if (
        (lbvc == 1)
        and str(stash) not in STASHCODE_IMPLIED_HEIGHTS
        and np.all(blev != -1)
    ):
        coord = _dim_or_aux(
            blev,
            standard_name="height",
            units="m",
            attributes={"positive": "up"},
        )
        coords_and_dims.append((coord, dim))

    if str(stash) in STASHCODE_IMPLIED_HEIGHTS:
        height = STASHCODE_IMPLIED_HEIGHTS[str(stash)]
        coord = DimCoord(
            height,
            standard_name="height",
            units="m",
            attributes={"positive": "up"},
        )
        coords_and_dims.append((coord, None))

    # Model level number.
    if (len(lbcode) != 5) and (lbvc == 2):
        coord = _dim_or_aux(
            model_level_number,
            standard_name="model_level_number",
            attributes={"positive": "down"},
            units="1",
        )
        coords_and_dims.append((coord, dim))

    # Depth - unbound.
    if (len(lbcode) != 5) and (lbvc == 2) and np.all(brsvd1 == brlev):
        coord = _dim_or_aux(
            blev,
            standard_name="depth",
            units="m",
            attributes={"positive": "down"},
        )
        coords_and_dims.append((coord, dim))

    # Depth - bound.
    if (len(lbcode) != 5) and (lbvc == 2) and np.all(brsvd1 != brlev):
        coord = _dim_or_aux(
            blev,
            standard_name="depth",
            units="m",
            bounds=np.vstack((brsvd1, brlev)).T,
            attributes={"positive": "down"},
        )
        coords_and_dims.append((coord, dim))

    # Depth - unbound and bound (mixed).
    if (
        (len(lbcode) != 5)
        and (lbvc == 2)
        and (np.any(brsvd1 == brlev) and np.any(brsvd1 != brlev))
    ):
        lower = np.where(brsvd1 == brlev, blev, brsvd1)
        upper = np.where(brsvd1 == brlev, blev, brlev)
        coord = _dim_or_aux(
            blev,
            standard_name="depth",
            units="m",
            bounds=np.vstack((lower, upper)).T,
            attributes={"positive": "down"},
        )
        coords_and_dims.append((coord, dim))

    # Soil level/depth.
    if len(lbcode) != 5 and lbvc == 6:
        if np.all(brsvd1 == 0) and np.all(brlev == 0):
            # UM populates lblev, brsvd1 and brlev metadata INCORRECTLY,
            # so continue to treat as a soil level.
            coord = _dim_or_aux(
                model_level_number,
                long_name="soil_model_level_number",
                attributes={"positive": "down"},
                units="1",
            )
            coords_and_dims.append((coord, dim))
        elif np.any(brsvd1 != brlev):
            # UM populates metadata CORRECTLY,
            # so treat it as the expected (bounded) soil depth.
            coord = _dim_or_aux(
                blev,
                standard_name="depth",
                units="m",
                bounds=np.vstack((brsvd1, brlev)).T,
                attributes={"positive": "down"},
            )
            coords_and_dims.append((coord, dim))

    # Pressure.
    if (lbvc == 8) and (
        len(lbcode) != 5
        or (len(lbcode) == 5 and 1 not in [lbcode.ix, lbcode.iy])
    ):
        coord = _dim_or_aux(blev, long_name="pressure", units="hPa")
        coords_and_dims.append((coord, dim))

    # Air potential temperature.
    if (len(lbcode) != 5) and (lbvc == 19):
        coord = _dim_or_aux(
            blev,
            standard_name="air_potential_temperature",
            units="K",
            attributes={"positive": "up"},
        )
        coords_and_dims.append((coord, dim))

    # Hybrid pressure levels.
    if lbvc == 9:
        model_level_number = _dim_or_aux(
            model_level_number,
            standard_name="model_level_number",
            attributes={"positive": "up"},
            units="1",
        )
        level_pressure = _dim_or_aux(
            bhlev,
            long_name="level_pressure",
            units="Pa",
            bounds=np.vstack((bhrlev, brsvd2)).T,
        )
        sigma = AuxCoord(
            blev,
            long_name="sigma",
            bounds=np.vstack((brlev, brsvd1)).T,
            units="1",
        )
        coords_and_dims.extend(
            [(model_level_number, dim), (level_pressure, dim), (sigma, dim)]
        )
        factories.append(
            Factory(
                HybridPressureFactory,
                [
                    {"long_name": "level_pressure"},
                    {"long_name": "sigma"},
                    Reference("surface_air_pressure"),
                ],
            )
        )

    # Hybrid height levels.
    if lbvc == 65:
        model_level_number = _dim_or_aux(
            model_level_number,
            standard_name="model_level_number",
            attributes={"positive": "up"},
            units="1",
        )
        level_height = _dim_or_aux(
            blev,
            long_name="level_height",
            units="m",
            bounds=np.vstack((brlev, brsvd1)).T,
            attributes={"positive": "up"},
        )
        sigma = AuxCoord(
            bhlev,
            long_name="sigma",
            bounds=np.vstack((bhrlev, brsvd2)).T,
            units="1",
        )
        coords_and_dims.extend(
            [(model_level_number, dim), (level_height, dim), (sigma, dim)]
        )
        factories.append(
            Factory(
                HybridHeightFactory,
                [
                    {"long_name": "level_height"},
                    {"long_name": "sigma"},
                    Reference("orography"),
                ],
            )
        )

    return coords_and_dims, factories


def _reshape_vector_args(values_and_dims):
    """
    Reshape a group of (array, dimensions-mapping) onto all dimensions.

    The resulting arrays are all mapped over the same dimensions; as many as
    the maximum dimension number found in the inputs.  Those dimensions not
    mapped by a given input appear as length-1 dimensions in the output array.
    The resulting arrays are thus all mutually compatible in arithmetic -- i.e.
    can combine without broadcasting errors (provided that all inputs mapping
    to a dimension define the same associated length).

    Args:

    * values_and_dims (iterable of (array-like, iterable of int)):
        Input arrays with associated mapping dimension numbers.
        The length of each 'dims' must match the ndims of the 'value'.

    Returns:

    * reshaped_arrays (iterable of arrays).
        The inputs, transposed and reshaped onto common target dimensions.

    """
    # Find maximum dimension index, which sets ndim of results.
    max_dims = [max(dims) if dims else -1 for _, dims in values_and_dims]
    max_dim = max(max_dims) if max_dims else -1
    result = []
    for value, dims in values_and_dims:
        value = np.asarray(value)
        if len(dims) != value.ndim:
            raise ValueError(
                "Lengths of dimension-mappings must match "
                "input array dimensions."
            )
        # Save dim sizes in original order.
        original_shape = value.shape
        if dims:
            # Transpose values to put its dims in the target order.
            dims_order = sorted(
                range(len(dims)), key=lambda i_dim: dims[i_dim]
            )
            value = value.transpose(dims_order)
        if max_dim != -1:
            # Reshape to add any extra *1 dims.
            shape = [1] * (max_dim + 1)
            for i_dim, dim in enumerate(dims):
                shape[dim] = original_shape[i_dim]
            value = value.reshape(shape)
        result.append(value)
    return result


def _collapse_degenerate_points_and_bounds(points, bounds=None, rtol=1.0e-7):
    """
    Collapse points (and optionally bounds) in any dimensions over which all
    values are the same.

    All dimensions are tested, and if degenerate are reduced to length 1.

    Value equivalence is controlled by a tolerance, to avoid problems with
    numbers from cftime.date2num, which has limited precision because of
    the way it calculates with floats of days.

    Args:

    * points (:class:`numpy.ndarray`)):
        Array of points values.

    Kwargs:

    * bounds (:class:`numpy.ndarray`)
        Array of bounds values. This array should have an additional vertex
        dimension (typically of length 2) when compared to the  points array
        i.e. bounds.shape = points.shape + (nvertex,)

    Returns:

        A (points, bounds) tuple.

    """
    array = points
    if bounds is not None:
        array = np.vstack((points, bounds.T)).T

    for i_dim in range(points.ndim):
        if array.shape[i_dim] > 1:
            slice_inds = [slice(None)] * points.ndim
            slice_inds[i_dim] = slice(0, 1)
            slice_0 = array[tuple(slice_inds)]
            if np.allclose(array, slice_0, rtol):
                array = slice_0

    points = array
    if bounds is not None:
        points = array[..., 0]
        bounds = array[..., 1:]

    return points, bounds


def _reduce_points_and_bounds(points, lower_and_upper_bounds=None):
    """
    Reduce the dimensionality of arrays of coordinate points (and optionally
    bounds).

    Dimensions over which all values are the same are reduced to size 1, using
    :func:`_collapse_degenerate_points_and_bounds`.
    All size-1 dimensions are then removed.
    If the bounds arrays are also passed in, then all three arrays must have
    the same shape or be capable of being broadcast to match.

    Args:

    * points (array-like):
        Coordinate point values.

    Kwargs:

    * lower_and_upper_bounds (pair of array-like, or None):
        Corresponding bounds values (lower, upper), if any.

    Returns:
        dims (iterable of ints), points(array), bounds(array)

        * 'dims' is the mapping from the result array dimensions to the
            original dimensions.  However, when 'array' is scalar, 'dims' will
            be None (rather than an empty tuple).
        * 'points' and 'bounds' are the reduced arrays.
            If no bounds were passed, None is returned.

    """
    orig_points_dtype = np.asarray(points).dtype
    bounds = None
    if lower_and_upper_bounds is not None:
        lower_bounds, upper_bounds = np.broadcast_arrays(
            *lower_and_upper_bounds
        )
        orig_bounds_dtype = lower_bounds.dtype
        bounds = np.vstack((lower_bounds, upper_bounds)).T

    # Attempt to broadcast points to match bounds to handle scalars.
    if bounds is not None and points.shape != bounds.shape[:-1]:
        points, _ = np.broadcast_arrays(points, bounds[..., 0])

    points, bounds = _collapse_degenerate_points_and_bounds(points, bounds)

    used_dims = tuple(
        i_dim for i_dim in range(points.ndim) if points.shape[i_dim] > 1
    )
    reshape_inds = tuple([points.shape[dim] for dim in used_dims])
    points = points.reshape(reshape_inds)
    points = points.astype(orig_points_dtype)
    if bounds is not None:
        bounds = bounds.reshape(reshape_inds + (2,))
        bounds = bounds.astype(orig_bounds_dtype)

    if not used_dims:
        used_dims = None

    return used_dims, points, bounds


def _new_coord_and_dims(
    is_vector_operation, name, units, points, lower_and_upper_bounds=None
):
    """
    Make a new (coordinate, cube_dims) pair with the given points, name, units
    and optional bounds.

    In 'vector' style operation, the data arrays must have same number of
    dimensions as the target cube, and additional operations are performed :

    * dimensions with all points and bounds values the same are removed.
    * the result coordinate may be an AuxCoord if a DimCoord cannot be made
        (e.g. if values are non-monotonic).

    Args:

    * is_vector_operation (bool):
        If True, perform 'vector' style operation.

    * points (array-like):
        Coordinate point values.

    * name (string):
        Standard name of coordinate.

    * units (string or cf_unit.Unit):
        Units of coordinate.

    Kwargs:

    * lower_and_upper_bounds (pair of array-like, or None):
        Corresponding bounds values (lower, upper), if any.

    Returns:
        a new (coordinate, dims) pair.

    """
    bounds = lower_and_upper_bounds
    if is_vector_operation:
        dims, points, bounds = _reduce_points_and_bounds(points, bounds)
    else:
        dims = None
    coord = _dim_or_aux(points, bounds=bounds, standard_name=name, units=units)
    return (coord, dims)


_HOURS_UNIT = cf_units.Unit("hours")


def _epoch_date_hours(epoch_hours_unit, datetime):
    """
    Return an 'hours since epoch' number for a date.

    Args:
    * epoch_hours_unit (:class:`cf_unit.Unit'):
        Unit defining the calendar and zero-time of conversion.
    * datetime (:class:`datetime.datetime`-like):
         Date object containing year / month / day attributes.

    This routine can also handle dates with a zero year, month or day  : such
    dates were valid inputs to 'date2num' up to cftime version 1.0.1, but are
    now illegal :  This routine interprets any zeros as being "1 year/month/day
    before a year/month/day of 1".  This produces results consistent with the
    "old" cftime behaviour.

    """
    days_offset = None
    if datetime.year == 0 or datetime.month == 0 or datetime.day == 0:
        # cftime > 1.0.1 no longer allows non-calendar dates.
        # Add 1 to year/month/day, to get a valid date, and adjust the result
        # according to the actual epoch and calendar.  This reproduces 'old'
        # results that were produced with cftime <= 1.0.1.
        days_offset = 0
        y, m, d = datetime.year, datetime.month, datetime.day
        calendar = epoch_hours_unit.calendar
        if d == 0:
            # Add one day, by changing day=0 to 1.
            d = 1
            days_offset += 1
        if m == 0:
            # Add a 'January', by changing month=0 to 1.
            m = 1
            if calendar == cf_units.CALENDAR_GREGORIAN:
                days_offset += 31
            elif calendar == cf_units.CALENDAR_360_DAY:
                days_offset += 30
            elif calendar == cf_units.CALENDAR_365_DAY:
                days_offset += 31
            else:
                msg = "unrecognised calendar : {}"
                raise ValueError(msg.format(calendar))

        if y == 0:
            # Add a 'Year 0', by changing year=0 to 1.
            y = 1
            if calendar == cf_units.CALENDAR_GREGORIAN:
                days_in_year_0 = 366
            elif calendar == cf_units.CALENDAR_360_DAY:
                days_in_year_0 = 360
            elif calendar == cf_units.CALENDAR_365_DAY:
                days_in_year_0 = 365
            else:
                msg = "unrecognised calendar : {}"
                raise ValueError(msg.format(calendar))

            days_offset += days_in_year_0

        # Replace y/m/d with a modified date, that cftime will accept.
        datetime = datetime.replace(year=y, month=m, day=d)

    # netcdf4python has changed it's behaviour, at version 1.2, such
    # that a date2num calculation returns a python float, not
    # numpy.float64.  The behaviour of round is to recast this to an
    # int, which is not the desired behaviour for PP files.
    # So, cast the answer to numpy.float_ to be safe.
    epoch_hours = np.float_(epoch_hours_unit.date2num(datetime))

    if days_offset is not None:
        # Correct for any modifications to achieve a valid date.
        epoch_hours -= 24.0 * days_offset

    return epoch_hours


def _convert_time_coords(
    lbcode,
    lbtim,
    epoch_hours_unit,
    t1,
    t2,
    lbft,
    t1_dims=(),
    t2_dims=(),
    lbft_dims=(),
):
    """
    Make time coordinates from the time metadata.

    Args:

    * lbcode(:class:`iris.fileformats.pp.SplittableInt`):
        Scalar field value.
    * lbtim (:class:`iris.fileformats.pp.SplittableInt`):
        Scalar field value.
    * epoch_hours_unit (:class:`cf_units.Unit`):
        Epoch time reference unit.
    * t1 (array-like or scalar):
        Scalar field value or an array of values.
    * t2 (array-like or scalar):
        Scalar field value or an array of values.
    * lbft (array-like or scalar):
        Scalar field value or an array of values.

    Kwargs:

    * t1_dims, t2_dims, lbft_dims (tuples of int):
        Cube dimension mappings for the array metadata. Each default to
        to (). The length of each dims tuple should equal the dimensionality
        of the corresponding array of values.

    Returns:

        A list of (coordinate, dims) tuples. The coordinates are instance of
        :class:`iris.coords.DimCoord` if possible, otherwise they are instance
        of :class:`iris.coords.AuxCoord`. When the coordinate is of length one,
        the `dims` value is None rather than an empty tuple.

    """

    def date2hours(t):
        epoch_hours = _epoch_date_hours(epoch_hours_unit, t)
        if t.minute == 0 and t.second == 0:
            epoch_hours = np.around(epoch_hours)
        return epoch_hours

    def date2year(t_in):
        return t_in.year

    # Check whether inputs are all scalar, for faster handling of scalar cases.
    do_vector = len(t1_dims) + len(t2_dims) + len(lbft_dims) > 0
    if do_vector:
        # Reform the input values so they have all the same number of
        # dimensions, transposing where necessary (based on the dimension
        # mappings) so that the dimensions are common across each array.
        # Note: this does not _guarantee_ that the arrays are broadcastable,
        # but subsequent arithmetic makes this assumption.
        t1, t2, lbft = _reshape_vector_args(
            [(t1, t1_dims), (t2, t2_dims), (lbft, lbft_dims)]
        )

        date2hours = np.vectorize(date2hours)
        date2year = np.vectorize(date2year)

    t1_epoch_hours = date2hours(t1)
    t2_epoch_hours = date2hours(t2)
    hours_from_t1_to_t2 = t2_epoch_hours - t1_epoch_hours
    hours_from_t2_to_t1 = t1_epoch_hours - t2_epoch_hours
    coords_and_dims = []

    if (
        (lbtim.ia == 0)
        and (lbtim.ib == 0)
        and (lbtim.ic in [1, 2, 3, 4])
        and (
            len(lbcode) != 5
            or (
                len(lbcode) == 5
                and lbcode.ix not in [20, 21, 22, 23]
                and lbcode.iy not in [20, 21, 22, 23]
            )
        )
    ):
        coords_and_dims.append(
            _new_coord_and_dims(
                do_vector, "time", epoch_hours_unit, t1_epoch_hours
            )
        )

    if (
        (lbtim.ia == 0)
        and (lbtim.ib == 1)
        and (lbtim.ic in [1, 2, 3, 4])
        and (
            len(lbcode) != 5
            or (
                len(lbcode) == 5
                and lbcode.ix not in [20, 21, 22, 23]
                and lbcode.iy not in [20, 21, 22, 23]
            )
        )
    ):
        coords_and_dims.append(
            _new_coord_and_dims(
                do_vector, "forecast_period", _HOURS_UNIT, hours_from_t2_to_t1
            )
        )
        coords_and_dims.append(
            _new_coord_and_dims(
                do_vector, "time", epoch_hours_unit, t1_epoch_hours
            )
        )
        coords_and_dims.append(
            _new_coord_and_dims(
                do_vector,
                "forecast_reference_time",
                epoch_hours_unit,
                t2_epoch_hours,
            )
        )

    if (
        (lbtim.ib == 2)
        and (lbtim.ic in [1, 2, 4])
        and (np.any(date2year(t1) != 0) and np.any(date2year(t2) != 0))
        and
        # Note: don't add time coordinates when years are zero and
        # lbtim.ib == 2.  These are handled elsewhere.
        (
            (len(lbcode) != 5)
            or (
                len(lbcode) == 5
                and lbcode.ix not in [20, 21, 22, 23]
                and lbcode.iy not in [20, 21, 22, 23]
            )
        )
    ):

        coords_and_dims.append(
            _new_coord_and_dims(
                do_vector,
                "forecast_period",
                _HOURS_UNIT,
                lbft - 0.5 * hours_from_t1_to_t2,
                [lbft - hours_from_t1_to_t2, lbft],
            )
        )

        coords_and_dims.append(
            _new_coord_and_dims(
                do_vector,
                "time",
                epoch_hours_unit,
                0.5 * (t1_epoch_hours + t2_epoch_hours),
                [t1_epoch_hours, t2_epoch_hours],
            )
        )

        coords_and_dims.append(
            _new_coord_and_dims(
                do_vector,
                "forecast_reference_time",
                epoch_hours_unit,
                t2_epoch_hours - lbft,
            )
        )

    if (
        (lbtim.ib == 3)
        and (lbtim.ic in [1, 2, 4])
        and (
            (len(lbcode) != 5)
            or (
                len(lbcode) == 5
                and lbcode.ix not in [20, 21, 22, 23]
                and lbcode.iy not in [20, 21, 22, 23]
            )
        )
    ):
        coords_and_dims.append(
            _new_coord_and_dims(
                do_vector,
                "forecast_period",
                _HOURS_UNIT,
                lbft,
                [lbft - hours_from_t1_to_t2, lbft],
            )
        )
        coords_and_dims.append(
            _new_coord_and_dims(
                do_vector,
                "time",
                epoch_hours_unit,
                t2_epoch_hours,
                [t1_epoch_hours, t2_epoch_hours],
            )
        )
        coords_and_dims.append(
            _new_coord_and_dims(
                do_vector,
                "forecast_reference_time",
                epoch_hours_unit,
                t2_epoch_hours - lbft,
            )
        )

    if (
        (len(lbcode) == 5)
        and (lbcode[-1] == 3)
        and (lbtim.ib == 2)
        and (lbtim.ic == 2)
    ):
        coords_and_dims.append(
            _new_coord_and_dims(
                do_vector,
                "forecast_reference_time",
                epoch_hours_unit,
                t2_epoch_hours - lbft,
            )
        )

    return coords_and_dims


###############################################################################


def _model_level_number(lblev):
    """
    Return model level number for an LBLEV value.

    Args:

    * lblev (int):
        PP field LBLEV value.

    Returns:
        Model level number (integer).

    """
    # See Word no. 33 (LBLEV) in section 4 of UM Model Docs (F3).
    SURFACE_AND_ZEROTH_RHO_LEVEL_LBLEV = 9999

    if lblev == SURFACE_AND_ZEROTH_RHO_LEVEL_LBLEV:
        model_level_number = 0
    else:
        model_level_number = lblev

    return model_level_number


def _convert_scalar_realization_coords(lbrsvd4):
    """
    Encode scalar 'realization' (aka ensemble) numbers as CM data.

    Returns a list of coords_and_dims.

    """
    # Realization (aka ensemble) (--> scalar coordinates)
    coords_and_dims = []
    if lbrsvd4 != 0:
        coords_and_dims.append(
            (DimCoord(lbrsvd4, standard_name="realization", units="1"), None)
        )
    return coords_and_dims


def _convert_scalar_pseudo_level_coords(lbuser5):
    """
    Encode scalar pseudo-level values as CM data.

    Returns a list of coords_and_dims.

    """
    coords_and_dims = []
    if lbuser5 != 0:
        coords_and_dims.append(
            (DimCoord(lbuser5, long_name="pseudo_level", units="1"), None)
        )
    return coords_and_dims


def convert(f):
    """
    Converts a PP field into the corresponding items of Cube metadata.

    Args:

    * f:
        A :class:`iris.fileformats.pp.PPField` object.

    Returns:
        A :class:`iris.fileformats.rules.ConversionMetadata` object.

    """
    factories = []
    aux_coords_and_dims = []

    # "Normal" (non-cross-sectional) Time values (--> scalar coordinates)
    time_coords_and_dims = _convert_time_coords(
        lbcode=f.lbcode,
        lbtim=f.lbtim,
        epoch_hours_unit=f.time_unit("hours"),
        t1=f.t1,
        t2=f.t2,
        lbft=f.lbft,
    )
    aux_coords_and_dims.extend(time_coords_and_dims)

    # "Normal" (non-cross-sectional) Vertical levels
    #    (--> scalar coordinates and factories)
    vertical_coords_and_dims, vertical_factories = _convert_vertical_coords(
        lbcode=f.lbcode,
        lbvc=f.lbvc,
        blev=f.blev,
        lblev=f.lblev,
        stash=f.stash,
        bhlev=f.bhlev,
        bhrlev=f.bhrlev,
        brsvd1=f.brsvd[0],
        brsvd2=f.brsvd[1],
        brlev=f.brlev,
    )
    aux_coords_and_dims.extend(vertical_coords_and_dims)
    factories.extend(vertical_factories)

    # Realization (aka ensemble) (--> scalar coordinates)
    aux_coords_and_dims.extend(
        _convert_scalar_realization_coords(lbrsvd4=f.lbrsvd[3])
    )

    # Pseudo-level coordinate (--> scalar coordinates)
    aux_coords_and_dims.extend(
        _convert_scalar_pseudo_level_coords(lbuser5=f.lbuser[4])
    )

    # All the other rules.
    (
        references,
        standard_name,
        long_name,
        units,
        attributes,
        cell_methods,
        dim_coords_and_dims,
        other_aux_coords_and_dims,
    ) = _all_other_rules(f)
    aux_coords_and_dims.extend(other_aux_coords_and_dims)

    return ConversionMetadata(
        factories,
        references,
        standard_name,
        long_name,
        units,
        attributes,
        cell_methods,
        dim_coords_and_dims,
        aux_coords_and_dims,
    )


def _all_other_rules(f):
    """
    This deals with all the other rules that have not been factored into any of
    the other convert_scalar_coordinate functions above.

    """
    references = []
    standard_name = None
    long_name = None
    units = None
    attributes = {}
    cell_methods = []
    dim_coords_and_dims = []
    aux_coords_and_dims = []

    # Season coordinates (--> scalar coordinates)
    if (
        f.lbtim.ib == 3
        and f.lbtim.ic in [1, 2, 4]
        and (
            len(f.lbcode) != 5
            or (
                len(f.lbcode) == 5
                and (
                    f.lbcode.ix not in [20, 21, 22, 23]
                    and f.lbcode.iy not in [20, 21, 22, 23]
                )
            )
        )
        and f.lbmon == 12
        and f.lbdat == 1
        and f.lbhr == 0
        and f.lbmin == 0
        and f.lbmond == 3
        and f.lbdatd == 1
        and f.lbhrd == 0
        and f.lbmind == 0
    ):
        aux_coords_and_dims.append(
            (AuxCoord("djf", long_name="season", units="no_unit"), None)
        )

    if (
        f.lbtim.ib == 3
        and f.lbtim.ic in [1, 2, 4]
        and (
            (len(f.lbcode) != 5)
            or (
                len(f.lbcode) == 5
                and f.lbcode.ix not in [20, 21, 22, 23]
                and f.lbcode.iy not in [20, 21, 22, 23]
            )
        )
        and f.lbmon == 3
        and f.lbdat == 1
        and f.lbhr == 0
        and f.lbmin == 0
        and f.lbmond == 6
        and f.lbdatd == 1
        and f.lbhrd == 0
        and f.lbmind == 0
    ):
        aux_coords_and_dims.append(
            (AuxCoord("mam", long_name="season", units="no_unit"), None)
        )

    if (
        f.lbtim.ib == 3
        and f.lbtim.ic in [1, 2, 4]
        and (
            (len(f.lbcode) != 5)
            or (
                len(f.lbcode) == 5
                and f.lbcode.ix not in [20, 21, 22, 23]
                and f.lbcode.iy not in [20, 21, 22, 23]
            )
        )
        and f.lbmon == 6
        and f.lbdat == 1
        and f.lbhr == 0
        and f.lbmin == 0
        and f.lbmond == 9
        and f.lbdatd == 1
        and f.lbhrd == 0
        and f.lbmind == 0
    ):
        aux_coords_and_dims.append(
            (AuxCoord("jja", long_name="season", units="no_unit"), None)
        )

    if (
        f.lbtim.ib == 3
        and f.lbtim.ic in [1, 2, 4]
        and (
            (len(f.lbcode) != 5)
            or (
                len(f.lbcode) == 5
                and f.lbcode.ix not in [20, 21, 22, 23]
                and f.lbcode.iy not in [20, 21, 22, 23]
            )
        )
        and f.lbmon == 9
        and f.lbdat == 1
        and f.lbhr == 0
        and f.lbmin == 0
        and f.lbmond == 12
        and f.lbdatd == 1
        and f.lbhrd == 0
        and f.lbmind == 0
    ):
        aux_coords_and_dims.append(
            (AuxCoord("son", long_name="season", units="no_unit"), None)
        )

    # Special case where year is zero and months match.
    # Month coordinates (--> scalar coordinates)
    if (
        f.lbtim.ib == 2
        and f.lbtim.ic in [1, 2, 4]
        and (
            (len(f.lbcode) != 5)
            or (
                len(f.lbcode) == 5
                and f.lbcode.ix not in [20, 21, 22, 23]
                and f.lbcode.iy not in [20, 21, 22, 23]
            )
        )
        and f.lbyr == 0
        and f.lbyrd == 0
        and f.lbmon == f.lbmond
    ):
        aux_coords_and_dims.append(
            (AuxCoord(f.lbmon, long_name="month_number", units="1"), None)
        )
        aux_coords_and_dims.append(
            (
                AuxCoord(
                    calendar.month_abbr[f.lbmon],
                    long_name="month",
                    units="no_unit",
                ),
                None,
            )
        )
        aux_coords_and_dims.append(
            (
                DimCoord(
                    points=f.lbft,
                    standard_name="forecast_period",
                    units="hours",
                ),
                None,
            )
        )

    # "Normal" (i.e. not cross-sectional) lats+lons (--> vector coordinates)
    if (
        f.bdx != 0.0
        and f.bdx != f.bmdi
        and len(f.lbcode) != 5
        and f.lbcode[0] == 1
    ):
        dim_coords_and_dims.append(
            (
                DimCoord.from_regular(
                    f.bzx,
                    f.bdx,
                    f.lbnpt,
                    standard_name=f._x_coord_name(),
                    units="degrees",
                    circular=(f.lbhem in [0, 4]),
                    coord_system=f.coord_system(),
                ),
                1,
            )
        )

    if (
        f.bdx != 0.0
        and f.bdx != f.bmdi
        and len(f.lbcode) != 5
        and f.lbcode[0] == 2
    ):
        dim_coords_and_dims.append(
            (
                DimCoord.from_regular(
                    f.bzx,
                    f.bdx,
                    f.lbnpt,
                    standard_name=f._x_coord_name(),
                    units="degrees",
                    circular=(f.lbhem in [0, 4]),
                    coord_system=f.coord_system(),
                    with_bounds=True,
                ),
                1,
            )
        )

    if (
        f.bdy != 0.0
        and f.bdy != f.bmdi
        and len(f.lbcode) != 5
        and f.lbcode[0] == 1
    ):
        dim_coords_and_dims.append(
            (
                DimCoord.from_regular(
                    f.bzy,
                    f.bdy,
                    f.lbrow,
                    standard_name=f._y_coord_name(),
                    units="degrees",
                    coord_system=f.coord_system(),
                ),
                0,
            )
        )

    if (
        f.bdy != 0.0
        and f.bdy != f.bmdi
        and len(f.lbcode) != 5
        and f.lbcode[0] == 2
    ):
        dim_coords_and_dims.append(
            (
                DimCoord.from_regular(
                    f.bzy,
                    f.bdy,
                    f.lbrow,
                    standard_name=f._y_coord_name(),
                    units="degrees",
                    coord_system=f.coord_system(),
                    with_bounds=True,
                ),
                0,
            )
        )

    if (f.bdy == 0.0 or f.bdy == f.bmdi) and (
        len(f.lbcode) != 5 or (len(f.lbcode) == 5 and f.lbcode.iy == 10)
    ):
        dim_coords_and_dims.append(
            (
                DimCoord(
                    f.y,
                    standard_name=f._y_coord_name(),
                    units="degrees",
                    bounds=f.y_bounds,
                    coord_system=f.coord_system(),
                ),
                0,
            )
        )

    if (f.bdx == 0.0 or f.bdx == f.bmdi) and (
        len(f.lbcode) != 5 or (len(f.lbcode) == 5 and f.lbcode.ix == 11)
    ):
        dim_coords_and_dims.append(
            (
                DimCoord(
                    f.x,
                    standard_name=f._x_coord_name(),
                    units="degrees",
                    bounds=f.x_bounds,
                    circular=(f.lbhem in [0, 4]),
                    coord_system=f.coord_system(),
                ),
                1,
            )
        )

    # Cross-sectional vertical level types (--> vector coordinates)
    if (
        len(f.lbcode) == 5
        and f.lbcode.iy == 2
        and (f.bdy == 0 or f.bdy == f.bmdi)
    ):
        dim_coords_and_dims.append(
            (
                DimCoord(
                    f.y,
                    standard_name="height",
                    units="km",
                    bounds=f.y_bounds,
                    attributes={"positive": "up"},
                ),
                0,
            )
        )

    if len(f.lbcode) == 5 and f.lbcode[-1] == 1 and f.lbcode.iy == 4:
        dim_coords_and_dims.append(
            (
                DimCoord(
                    f.y,
                    standard_name="depth",
                    units="m",
                    bounds=f.y_bounds,
                    attributes={"positive": "down"},
                ),
                0,
            )
        )

    if (
        len(f.lbcode) == 5
        and f.lbcode.ix == 10
        and f.bdx != 0
        and f.bdx != f.bmdi
    ):
        dim_coords_and_dims.append(
            (
                DimCoord.from_regular(
                    f.bzx,
                    f.bdx,
                    f.lbnpt,
                    standard_name=f._y_coord_name(),
                    units="degrees",
                    coord_system=f.coord_system(),
                ),
                1,
            )
        )

    if (
        len(f.lbcode) == 5
        and f.lbcode.iy == 1
        and (f.bdy == 0 or f.bdy == f.bmdi)
    ):
        dim_coords_and_dims.append(
            (
                DimCoord(
                    f.y, long_name="pressure", units="hPa", bounds=f.y_bounds
                ),
                0,
            )
        )

    if (
        len(f.lbcode) == 5
        and f.lbcode.ix == 1
        and (f.bdx == 0 or f.bdx == f.bmdi)
    ):
        dim_coords_and_dims.append(
            (
                DimCoord(
                    f.x, long_name="pressure", units="hPa", bounds=f.x_bounds
                ),
                1,
            )
        )

    # Cross-sectional time values (--> vector coordinates)
    if len(f.lbcode) == 5 and f.lbcode[-1] == 1 and f.lbcode.iy == 23:
        dim_coords_and_dims.append(
            (
                DimCoord(
                    f.y,
                    standard_name="time",
                    units=cf_units.Unit(
                        "days since 0000-01-01 00:00:00",
                        calendar=cf_units.CALENDAR_360_DAY,
                    ),
                    bounds=f.y_bounds,
                ),
                0,
            )
        )

    if len(f.lbcode) == 5 and f.lbcode[-1] == 1 and f.lbcode.ix == 23:
        dim_coords_and_dims.append(
            (
                DimCoord(
                    f.x,
                    standard_name="time",
                    units=cf_units.Unit(
                        "days since 0000-01-01 00:00:00",
                        calendar=cf_units.CALENDAR_360_DAY,
                    ),
                    bounds=f.x_bounds,
                ),
                1,
            )
        )

    if (
        len(f.lbcode) == 5
        and f.lbcode[-1] == 3
        and f.lbcode.iy == 23
        and f.lbtim.ib == 2
        and f.lbtim.ic == 2
    ):
        epoch_days_unit = cf_units.Unit(
            "days since 0000-01-01 00:00:00",
            calendar=cf_units.CALENDAR_360_DAY,
        )
        t1_epoch_days = epoch_days_unit.date2num(f.t1)
        t2_epoch_days = epoch_days_unit.date2num(f.t2)
        # The end time is exclusive, not inclusive.
        dim_coords_and_dims.append(
            (
                DimCoord(
                    np.linspace(
                        t1_epoch_days, t2_epoch_days, f.lbrow, endpoint=False
                    ),
                    standard_name="time",
                    units=epoch_days_unit,
                    bounds=f.y_bounds,
                ),
                0,
            )
        )

    # Site number (--> scalar coordinate)
    if (
        len(f.lbcode) == 5
        and f.lbcode[-1] == 1
        and f.lbcode.ix == 13
        and f.bdx != 0
    ):
        dim_coords_and_dims.append(
            (
                DimCoord.from_regular(
                    f.bzx, f.bdx, f.lbnpt, long_name="site_number", units="1"
                ),
                1,
            )
        )

    # Site number cross-sections (???)
    if (
        len(f.lbcode) == 5
        and 13 in [f.lbcode.ix, f.lbcode.iy]
        and 11 not in [f.lbcode.ix, f.lbcode.iy]
        and hasattr(f, "lower_x_domain")
        and hasattr(f, "upper_x_domain")
        and all(f.lower_x_domain != -1.0e30)
        and all(f.upper_x_domain != -1.0e30)
    ):
        aux_coords_and_dims.append(
            (
                AuxCoord(
                    (f.lower_x_domain + f.upper_x_domain) / 2.0,
                    standard_name=f._x_coord_name(),
                    units="degrees",
                    bounds=np.array([f.lower_x_domain, f.upper_x_domain]).T,
                    coord_system=f.coord_system(),
                ),
                1 if f.lbcode.ix == 13 else 0,
            )
        )

    if (
        len(f.lbcode) == 5
        and 13 in [f.lbcode.ix, f.lbcode.iy]
        and 10 not in [f.lbcode.ix, f.lbcode.iy]
        and hasattr(f, "lower_y_domain")
        and hasattr(f, "upper_y_domain")
        and all(f.lower_y_domain != -1.0e30)
        and all(f.upper_y_domain != -1.0e30)
    ):
        aux_coords_and_dims.append(
            (
                AuxCoord(
                    (f.lower_y_domain + f.upper_y_domain) / 2.0,
                    standard_name=f._y_coord_name(),
                    units="degrees",
                    bounds=np.array([f.lower_y_domain, f.upper_y_domain]).T,
                    coord_system=f.coord_system(),
                ),
                1 if f.lbcode.ix == 13 else 0,
            )
        )

    # LBPROC codings (--> cell method + attributes)
    unhandled_lbproc = True
    zone_method = None
    time_method = None
    if f.lbproc == 0:
        unhandled_lbproc = False
    elif f.lbproc == 64:
        zone_method = "mean"
    elif f.lbproc == 128:
        time_method = "mean"
    elif f.lbproc == 4096:
        time_method = "minimum"
    elif f.lbproc == 8192:
        time_method = "maximum"
    elif f.lbproc == 192:
        time_method = "mean"
        zone_method = "mean"

    if time_method is not None:
        if f.lbtim.ia != 0:
            intervals = "{} hour".format(f.lbtim.ia)
        else:
            intervals = None

        if f.lbtim.ib == 2:
            # Aggregation over a period of time.
            cell_methods.append(
                CellMethod(time_method, coords="time", intervals=intervals)
            )
            unhandled_lbproc = False
        elif f.lbtim.ib == 3 and f.lbproc == 128:
            # Aggregation over a period of time within a year, over a number
            # of years.
            # Only mean (lbproc of 128) is handled as the min/max
            # interpretation is ambiguous e.g. decadal mean of daily max,
            # decadal max of daily mean, decadal mean of max daily mean etc.
            cell_methods.append(
                CellMethod(
                    "{} within years".format(time_method),
                    coords="time",
                    intervals=intervals,
                )
            )
            cell_methods.append(
                CellMethod("{} over years".format(time_method), coords="time")
            )
            unhandled_lbproc = False
        else:
            # Generic cell method to indicate a time aggregation.
            cell_methods.append(CellMethod(time_method, coords="time"))
            unhandled_lbproc = False

    if zone_method is not None:
        if f.lbcode == 1:
            cell_methods.append(CellMethod(zone_method, coords="longitude"))
            for coord, _dim in dim_coords_and_dims:
                if coord.standard_name == "longitude":
                    if len(coord.points) == 1:
                        coord.bounds = np.array([0.0, 360.0], dtype=np.float32)
                    else:
                        coord.guess_bounds()
            unhandled_lbproc = False
        elif f.lbcode == 101:
            cell_methods.append(
                CellMethod(zone_method, coords="grid_longitude")
            )
            for coord, _dim in dim_coords_and_dims:
                if coord.standard_name == "grid_longitude":
                    if len(coord.points) == 1:
                        coord.bounds = np.array([0.0, 360.0], dtype=np.float32)
                    else:
                        coord.guess_bounds()
            unhandled_lbproc = False
        else:
            unhandled_lbproc = True

    if unhandled_lbproc:
        attributes["ukmo__process_flags"] = tuple(
            sorted(
                [
                    name
                    for value, name in LBPROC_MAP.items()
                    if isinstance(value, int) and f.lbproc & value
                ]
            )
        )

    if (f.lbsrce % 10000) == 1111:
        attributes["source"] = "Data from Met Office Unified Model"
        # Also define MO-netCDF compliant UM version.
        um_major = (f.lbsrce // 10000) // 100
        if um_major != 0:
            um_minor = (f.lbsrce // 10000) % 100
            attributes["um_version"] = "{:d}.{:d}".format(um_major, um_minor)

    if (
        f.lbuser[6] != 0
        or (f.lbuser[3] // 1000) != 0
        or (f.lbuser[3] % 1000) != 0
    ):
        attributes["STASH"] = f.stash

    if str(f.stash) in STASH_TO_CF:
        standard_name = STASH_TO_CF[str(f.stash)].standard_name
        units = STASH_TO_CF[str(f.stash)].units
        long_name = STASH_TO_CF[str(f.stash)].long_name

    if not f.stash.is_valid and f.lbfc in LBFC_TO_CF:
        standard_name = LBFC_TO_CF[f.lbfc].standard_name
        units = LBFC_TO_CF[f.lbfc].units
        long_name = LBFC_TO_CF[f.lbfc].long_name

    # Orography reference field (--> reference target)
    if f.lbuser[3] == 33:
        references.append(ReferenceTarget("orography", None))

    # Surface pressure reference field (--> reference target)
    if f.lbuser[3] == 409 or f.lbuser[3] == 1:
        references.append(ReferenceTarget("surface_air_pressure", None))

    return (
        references,
        standard_name,
        long_name,
        units,
        attributes,
        cell_methods,
        dim_coords_and_dims,
        aux_coords_and_dims,
    )



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Test coordinate categorisation function add_hour.
"""

# import iris tests first so that some things can be initialised before
# importing anything else
import iris.tests as tests

import cf_units
import numpy as np

import iris
import iris.coord_categorisation as ccat


class Test_add_hour(tests.IrisTest):
    def setUp(self):
        # make a series of 'hour numbers' for the time
        hour_numbers = np.arange(0, 200, 5, dtype=np.int32)

        # use hour numbers as data values also (don't actually use this for
        # anything)
        cube = iris.cube.Cube(
            hour_numbers, long_name="test cube", units="metres"
        )

        time_coord = iris.coords.DimCoord(
            hour_numbers,
            standard_name="time",
            units=cf_units.Unit("hours since epoch", "gregorian"),
        )
        cube.add_dim_coord(time_coord, 0)

        self.hour_numbers = hour_numbers
        self.cube = cube
        self.time_coord = time_coord

    def test_bad_coord(self):
        with self.assertRaises(iris.exceptions.CoordinateNotFoundError):
            ccat.add_hour(self.cube, "DOES NOT EXIST", name="my_hour")

    def test_explicit_result_name_specify_coord_by_name(self):
        coord_name = "my_hour"
        msg = "Missing/incorrectly named result for add_hour"

        # Specify source coordinate by name
        cube = self.cube
        ccat.add_hour(cube, "time", name=coord_name)
        result_coords = cube.coords(coord_name)
        self.assertEqual(len(result_coords), 1, msg)

    def test_explicit_result_name_specify_coord_by_reference(self):
        coord_name = "my_hour"
        msg = "Missing/incorrectly named result for add_hour"

        # Specify source coordinate by coordinate reference
        cube = self.cube
        time = cube.coord("time")
        ccat.add_hour(cube, time, name=coord_name)
        result_coords = cube.coords(coord_name)
        self.assertEqual(len(result_coords), 1, msg)

    def test_basic(self):
        coord_name = "my_hour"
        cube = self.cube
        time_coord = self.time_coord
        expected_coord = iris.coords.AuxCoord(
            self.hour_numbers % 24, long_name=coord_name, units="1"
        )

        ccat.add_hour(cube, time_coord, coord_name)

        self.assertEqual(cube.coord(coord_name), expected_coord)


if __name__ == "__main__":
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""Unit tests for the `iris._merge.ProtoCube` class."""

# Import iris.tests first so that some things can be initialised before
# importing anything else.
import iris.tests as tests

from abc import ABCMeta, abstractmethod
from unittest import mock

import numpy as np
import numpy.ma as ma

import iris
from iris._merge import ProtoCube
from iris.aux_factory import HybridHeightFactory, HybridPressureFactory
from iris.coords import DimCoord, AuxCoord
from iris.exceptions import MergeError


def example_cube():
    return iris.cube.Cube(
        np.array([1, 2, 3], dtype="i4"),
        standard_name="air_temperature",
        long_name="screen_air_temp",
        var_name="airtemp",
        units="K",
        attributes={"mint": "thin"},
    )


class Mixin_register(metaclass=ABCMeta):
    @property
    def cube1(self):
        return example_cube()

    @property
    @abstractmethod
    def cube2(self):
        pass

    @property
    @abstractmethod
    def fragments(self):
        pass

    def test_default(self):
        # Test what happens when we call:
        #   ProtoCube.register(cube)
        proto_cube = ProtoCube(self.cube1)
        result = proto_cube.register(self.cube2)
        self.assertEqual(result, not self.fragments)

    def test_no_error(self):
        # Test what happens when we call:
        #   ProtoCube.register(cube, error_on_mismatch=False)
        proto_cube = ProtoCube(self.cube1)
        result = proto_cube.register(self.cube2, error_on_mismatch=False)
        self.assertEqual(result, not self.fragments)

    def test_error(self):
        # Test what happens when we call:
        #   ProtoCube.register(cube, error_on_mismatch=True)
        proto_cube = ProtoCube(self.cube1)
        if self.fragments:
            with self.assertRaises(iris.exceptions.MergeError) as cm:
                proto_cube.register(self.cube2, error_on_mismatch=True)
            error_message = str(cm.exception)
            for substr in self.fragments:
                self.assertIn(substr, error_message)
        else:
            result = proto_cube.register(self.cube2, error_on_mismatch=True)
            self.assertTrue(result)


@tests.iristest_timing_decorator
class Test_register__match(Mixin_register, tests.IrisTest_nometa):
    @property
    def fragments(self):
        return []

    @property
    def cube2(self):
        return example_cube()


@tests.iristest_timing_decorator
class Test_register__standard_name(Mixin_register, tests.IrisTest_nometa):
    @property
    def fragments(self):
        return ["cube.standard_name", "air_temperature", "air_density"]

    @property
    def cube2(self):
        cube = example_cube()
        cube.standard_name = "air_density"
        return cube


@tests.iristest_timing_decorator
class Test_register__long_name(Mixin_register, tests.IrisTest_nometa):
    @property
    def fragments(self):
        return ["cube.long_name", "screen_air_temp", "Belling"]

    @property
    def cube2(self):
        cube = example_cube()
        cube.long_name = "Belling"
        return cube


@tests.iristest_timing_decorator
class Test_register__var_name(Mixin_register, tests.IrisTest_nometa):
    @property
    def fragments(self):
        return ["cube.var_name", "'airtemp'", "'airtemp2'"]

    @property
    def cube2(self):
        cube = example_cube()
        cube.var_name = "airtemp2"
        return cube


@tests.iristest_timing_decorator
class Test_register__units(Mixin_register, tests.IrisTest_nometa):
    @property
    def fragments(self):
        return ["cube.units", "'K'", "'C'"]

    @property
    def cube2(self):
        cube = example_cube()
        cube.units = "C"
        return cube


@tests.iristest_timing_decorator
class Test_register__attributes_unequal(Mixin_register, tests.IrisTest_nometa):
    @property
    def fragments(self):
        return ["cube.attributes", "'mint'"]

    @property
    def cube2(self):
        cube = example_cube()
        cube.attributes["mint"] = "waffer-thin"
        return cube


@tests.iristest_timing_decorator
class Test_register__attributes_unequal_array(
    Mixin_register, tests.IrisTest_nometa
):
    @property
    def fragments(self):
        return ["cube.attributes", "'mint'"]

    @property
    def cube1(self):
        cube = example_cube()
        cube.attributes["mint"] = np.arange(3)
        return cube

    @property
    def cube2(self):
        cube = example_cube()
        cube.attributes["mint"] = np.arange(3) + 1
        return cube


@tests.iristest_timing_decorator
class Test_register__attributes_superset(
    Mixin_register, tests.IrisTest_nometa
):
    @property
    def fragments(self):
        return ["cube.attributes", "'stuffed'"]

    @property
    def cube2(self):
        cube = example_cube()
        cube.attributes["stuffed"] = "yes"
        return cube


@tests.iristest_timing_decorator
class Test_register__attributes_multi_diff(
    Mixin_register, tests.IrisTest_nometa
):
    @property
    def fragments(self):
        return ["cube.attributes", "'sam'", "'mint'"]

    @property
    def cube1(self):
        cube = example_cube()
        cube.attributes["ralph"] = 1
        cube.attributes["sam"] = 2
        cube.attributes["tom"] = 3
        return cube

    @property
    def cube2(self):
        cube = example_cube()
        cube.attributes["ralph"] = 1
        cube.attributes["sam"] = "mug"
        cube.attributes["tom"] = 3
        cube.attributes["mint"] = "humbug"
        return cube


@tests.iristest_timing_decorator
class Test_register__cell_method(Mixin_register, tests.IrisTest_nometa):
    @property
    def fragments(self):
        return ["cube.cell_methods"]

    @property
    def cube2(self):
        cube = example_cube()
        cube.add_cell_method(iris.coords.CellMethod("monty", ("python",)))
        return cube


@tests.iristest_timing_decorator
class Test_register__data_shape(Mixin_register, tests.IrisTest_nometa):
    @property
    def fragments(self):
        return ["cube.shape", "(2,)", "(3,)"]

    @property
    def cube2(self):
        cube = example_cube()
        cube = cube[1:]
        return cube


@tests.iristest_timing_decorator
class Test_register__data_dtype(Mixin_register, tests.IrisTest_nometa):
    @property
    def fragments(self):
        return ["cube data dtype", "int32", "int8"]

    @property
    def cube2(self):
        cube = example_cube()
        cube.data = cube.data.astype(np.int8)
        return cube


class _MergeTest:
    # A mixin test class for common test methods implementation.

    # used by check routine: inheritors must implement it
    _mergetest_type = NotImplementedError

    def check_merge_fails_with_message(self):
        proto_cube = iris._merge.ProtoCube(self.cube1)
        with self.assertRaises(MergeError) as arc:
            proto_cube.register(self.cube2, error_on_mismatch=True)
        return str(arc.exception)

    def check_fail(self, *substrs):
        if isinstance(substrs, str):
            substrs = [substrs]
        msg = self.check_merge_fails_with_message()
        for substr in substrs:
            self.assertIn(substr, msg)


class Test_register__CubeSig(_MergeTest, tests.IrisTest):
    # Test potential registration failures.

    _mergetest_type = "cube"

    def setUp(self):
        self.cube1 = iris.cube.Cube(
            [1, 2, 3],
            standard_name="air_temperature",
            units="K",
            attributes={"mint": "thin"},
        )
        self.cube2 = self.cube1.copy()

    def test_noise(self):
        # Test a massive set of all defn diffs to make sure it's not noise.
        self.cube1.var_name = "Arthur"
        cube2 = self.cube1[1:]
        cube2.data = cube2.data.astype(np.int8)
        cube2.data = ma.array(cube2.data)
        cube2.standard_name = "air_pressure"
        cube2.var_name = "Nudge"
        cube2.attributes["stuffed"] = "yes"
        cube2.attributes["mint"] = "waffer-thin"
        cube2.add_cell_method(iris.coords.CellMethod("monty", ("python",)))

        # Check the actual message, so we've got a readable reference text.
        self.cube2 = cube2
        msg = self.check_merge_fails_with_message()
        self.assertString(msg, self.result_path(ext="txt"))


class Test_register__CoordSig_general(_MergeTest, tests.IrisTest):

    _mergetest_type = "coord"

    def setUp(self):
        self.cube1 = iris.cube.Cube(np.zeros((3, 3, 3)))
        self.cube2 = self.cube1.copy()

    def test_scalar_defns_one_extra(self):
        self.cube2.add_aux_coord(DimCoord([1], standard_name="latitude"))
        self.check_fail("aux_coords (scalar)", "latitude")

    def test_scalar_defns_both_extra(self):
        self.cube2.add_aux_coord(DimCoord([1], standard_name="latitude"))
        self.cube1.add_aux_coord(DimCoord([1], standard_name="longitude"))
        self.check_fail("aux_coords (scalar)", "latitude", "longitude")

    def test_vector_dim_coords_and_dims_one_extra(self):
        self.cube2.add_dim_coord(
            DimCoord([1, 2, 3], standard_name="latitude"), 0
        )
        self.check_fail("dim_coords", "latitude")

    def test_vector_dim_coords_and_dims_both_extra(self):
        self.cube2.add_dim_coord(
            DimCoord([1, 2, 3], standard_name="latitude"), 0
        )
        self.cube1.add_dim_coord(
            DimCoord([1, 2, 3], standard_name="longitude"), 0
        )
        self.check_fail("dim_coords", "latitude", "longitude")

    def test_vector_aux_coords_and_dims_one_extra(self):
        self.cube2.add_aux_coord(
            DimCoord([1, 2, 3], standard_name="latitude"), 0
        )
        self.check_fail("aux_coords (non-scalar)", "latitude")

    def test_vector_aux_coords_and_dims_both_extra(self):
        self.cube2.add_aux_coord(
            DimCoord([1, 2, 3], standard_name="latitude"), 0
        )
        self.cube1.add_aux_coord(
            DimCoord([1, 2, 3], standard_name="longitude"), 0
        )
        self.check_fail("aux_coords (non-scalar)", "latitude", "longitude")

    def test_factory_defns_one_extra(self):
        self.cube2.add_aux_factory(mock.MagicMock(spec=HybridHeightFactory))
        self.check_fail("cube.aux_factories", "differ")

    def test_factory_defns_both_extra(self):
        self.cube2.add_aux_factory(mock.MagicMock(spec=HybridHeightFactory))
        self.cube1.add_aux_factory(mock.MagicMock(spec=HybridPressureFactory))
        self.check_fail("cube.aux_factories", "differ")

    def test_factory_defns_one_missing_term(self):
        self.cube1.add_aux_factory(mock.MagicMock(spec=HybridPressureFactory))
        no_delta_factory = mock.MagicMock(spec=HybridPressureFactory)
        no_delta_factory.delta = None
        self.cube2.add_aux_factory(no_delta_factory)

        self.check_fail("cube.aux_factories", "differ")

    def test_noise(self):
        cube2 = self.cube2

        # scalar
        cube2.add_aux_coord(DimCoord([1], long_name="liff"))
        cube2.add_aux_coord(DimCoord([1], long_name="life"))
        cube2.add_aux_coord(DimCoord([1], long_name="like"))

        self.cube1.add_aux_coord(DimCoord([1], var_name="ming"))
        self.cube1.add_aux_coord(DimCoord([1], var_name="mong"))
        self.cube1.add_aux_coord(DimCoord([1], var_name="moog"))

        # aux
        cube2.add_dim_coord(DimCoord([1, 2, 3], standard_name="latitude"), 0)
        cube2.add_dim_coord(DimCoord([1, 2, 3], standard_name="longitude"), 1)
        cube2.add_dim_coord(DimCoord([1, 2, 3], standard_name="altitude"), 2)

        self.cube1.add_dim_coord(
            DimCoord([1, 2, 3], long_name="equinimity"), 0
        )
        self.cube1.add_dim_coord(
            DimCoord([1, 2, 3], long_name="equinomity"), 1
        )
        self.cube1.add_dim_coord(
            DimCoord([1, 2, 3], long_name="equinumity"), 2
        )

        # dim
        cube2.add_aux_coord(DimCoord([1, 2, 3], var_name="one"), 0)
        cube2.add_aux_coord(DimCoord([1, 2, 3], var_name="two"), 1)
        cube2.add_aux_coord(DimCoord([1, 2, 3], var_name="three"), 2)

        self.cube1.add_aux_coord(DimCoord([1, 2, 3], long_name="ay"), 0)
        self.cube1.add_aux_coord(DimCoord([1, 2, 3], long_name="bee"), 1)
        self.cube1.add_aux_coord(DimCoord([1, 2, 3], long_name="cee"), 2)

        # factory
        cube2.add_aux_factory(mock.MagicMock(spec=HybridHeightFactory))
        self.cube1.add_aux_factory(mock.MagicMock(spec=HybridPressureFactory))

        # Check the actual message, so we've got a readable reference text.
        self.cube2 = cube2
        msg = self.check_merge_fails_with_message()
        self.assertString(msg, self.result_path(ext="txt"))


class _MergeTest_coordprops(_MergeTest):
    # A mixin test class for common coordinate properties tests.

    # This must be implemented by inheritors.
    _mergetest_type = NotImplementedError

    def test_nochange(self):
        # This should simply succeed.
        proto_cube = iris._merge.ProtoCube(self.cube1)
        proto_cube.register(self.cube2, error_on_mismatch=True)

    def _props_fail(self, *terms):
        self.check_fail(
            self._mergetest_type, self.coord_to_change.name(), *terms
        )

    def test_standard_name(self):
        self.coord_to_change.standard_name = "soil_temperature"
        self._props_fail("air_temperature", "soil_temperature")

    def test_long_name(self):
        self.coord_to_change.long_name = "alternate_name"
        self._props_fail("air_temperature")

    def test_var_name(self):
        self.coord_to_change.var_name = "alternate_name"
        self._props_fail("air_temperature")

    def test_units(self):
        self.coord_to_change.units = "m"
        self._props_fail("air_temperature")

    def test_attrs_unequal(self):
        self.coord_to_change.attributes["att_a"] = 99
        self._props_fail("air_temperature")

    def test_attrs_set(self):
        self.coord_to_change.attributes["att_extra"] = 101
        self._props_fail("air_temperature")

    def test_coord_system(self):
        self.coord_to_change.coord_system = mock.Mock()
        self._props_fail("air_temperature")


class Test_register__CoordSig_scalar(_MergeTest_coordprops, tests.IrisTest):

    _mergetest_type = "aux_coords (scalar)"

    def setUp(self):
        self.cube1 = iris.cube.Cube(np.zeros((3, 3, 3)))
        self.cube1.add_aux_coord(
            DimCoord(
                [1],
                standard_name="air_temperature",
                long_name="eg_scalar",
                var_name="t1",
                units="K",
                attributes={"att_a": 1, "att_b": 2},
                coord_system=None,
            )
        )
        self.coord_to_change = self.cube1.coord("air_temperature")
        self.cube2 = self.cube1.copy()


class _MergeTest_coordprops_vect(_MergeTest_coordprops):
    # A derived mixin test class.
    # Adds extra props test for aux+dim coords (test points, bounds + dims)
    _mergetest_type = NotImplementedError
    _coord_typename = NotImplementedError

    def test_points(self):
        self.coord_to_change.points = self.coord_to_change.points + 1.0
        self.check_fail(self._mergetest_type, "air_temperature")

    def test_bounds(self):
        self.coord_to_change.bounds = self.coord_to_change.bounds + 1.0
        self.check_fail(self._mergetest_type, "air_temperature")

    def test_dims(self):
        self.cube2.remove_coord(self.coord_to_change)
        cube2_add_method = getattr(self.cube2, "add_" + self._coord_typename)
        cube2_add_method(self.coord_to_change, (1,))
        self.check_fail(self._mergetest_type, "mapping")


class Test_register__CoordSig_dim(_MergeTest_coordprops_vect, tests.IrisTest):

    _mergetest_type = "dim_coords"
    _coord_typename = "dim_coord"

    def setUp(self):
        self.cube1 = iris.cube.Cube(np.zeros((3, 3)))
        self.cube1.add_dim_coord(
            DimCoord(
                [15, 25, 35],
                bounds=[[10, 20], [20, 30], [30, 40]],
                standard_name="air_temperature",
                long_name="eg_scalar",
                var_name="t1",
                units="K",
                attributes={"att_a": 1, "att_b": 2},
                coord_system=None,
            ),
            (0,),
        )
        self.coord_to_change = self.cube1.coord("air_temperature")
        self.cube2 = self.cube1.copy()

    def test_circular(self):
        # Extra failure mode that only applies to dim coords
        self.coord_to_change.circular = True
        self.check_fail(self._mergetest_type, "air_temperature")


class Test_register__CoordSig_aux(_MergeTest_coordprops_vect, tests.IrisTest):

    _mergetest_type = "aux_coords (non-scalar)"
    _coord_typename = "aux_coord"

    def setUp(self):
        self.cube1 = iris.cube.Cube(np.zeros((3, 3)))
        self.cube1.add_aux_coord(
            AuxCoord(
                [65, 45, 85],
                bounds=[[60, 70], [40, 50], [80, 90]],
                standard_name="air_temperature",
                long_name="eg_scalar",
                var_name="t1",
                units="K",
                attributes={"att_a": 1, "att_b": 2},
                coord_system=None,
            ),
            (0,),
        )
        self.coord_to_change = self.cube1.coord("air_temperature")
        self.cube2 = self.cube1.copy()


if __name__ == "__main__":
    tests.main()



"""
Plotting in Different Projections
=================================

This example shows how to overlay data and graphics in different projections,
demonstrating various features of Iris, Cartopy and matplotlib.

We wish to overlay two datasets, defined on different rotated-pole grids.
To display both together, we make a pseudocoloured plot of the first, overlaid
with contour lines from the second.
We also add some lines and text annotations drawn in various projections.

We plot these over a specified region, in two different map projections.

"""

import cartopy.crs as ccrs
import matplotlib.pyplot as plt
import numpy as np

import iris
import iris.plot as iplt


# Define a Cartopy 'ordinary' lat-lon coordinate reference system.
crs_latlon = ccrs.PlateCarree()


def make_plot(projection_name, projection_crs):

    # Create a matplotlib Figure.
    plt.figure()

    # Add a matplotlib Axes, specifying the required display projection.
    # NOTE: specifying 'projection' (a "cartopy.crs.Projection") makes the
    # resulting Axes a "cartopy.mpl.geoaxes.GeoAxes", which supports plotting
    # in different coordinate systems.
    ax = plt.axes(projection=projection_crs)

    # Set display limits to include a set region of latitude * longitude.
    # (Note: Cartopy-specific).
    ax.set_extent((-80.0, 20.0, 10.0, 80.0), crs=crs_latlon)

    # Add coastlines and meridians/parallels (Cartopy-specific).
    ax.coastlines(linewidth=0.75, color="navy")
    ax.gridlines(crs=crs_latlon, linestyle="-")

    # Plot the first dataset as a pseudocolour filled plot.
    maindata_filepath = iris.sample_data_path("rotated_pole.nc")
    main_data = iris.load_cube(maindata_filepath)
    # NOTE: iplt.pcolormesh calls "pyplot.pcolormesh", passing in a coordinate
    # system with the 'transform' keyword:  This enables the Axes (a cartopy
    # GeoAxes) to reproject the plot into the display projection.
    iplt.pcolormesh(main_data, cmap="RdBu_r")

    # Overplot the other dataset (which has a different grid), as contours.
    overlay_filepath = iris.sample_data_path("space_weather.nc")
    overlay_data = iris.load_cube(overlay_filepath, "total electron content")
    # NOTE: as above, "iris.plot.contour" calls "pyplot.contour" with a
    # 'transform' keyword, enabling Cartopy reprojection.
    iplt.contour(
        overlay_data, 20, linewidths=2.0, colors="darkgreen", linestyles="-"
    )

    # Draw a margin line, some way in from the border of the 'main' data...
    # First calculate rectangle corners, 7% in from each corner of the data.
    x_coord, y_coord = main_data.coord(axis="x"), main_data.coord(axis="y")
    x_start, x_end = np.min(x_coord.points), np.max(x_coord.points)
    y_start, y_end = np.min(y_coord.points), np.max(y_coord.points)
    margin = 0.07
    margin_fractions = np.array([margin, 1.0 - margin])
    x_lower, x_upper = x_start + (x_end - x_start) * margin_fractions
    y_lower, y_upper = y_start + (y_end - y_start) * margin_fractions
    box_x_points = x_lower + (x_upper - x_lower) * np.array([0, 1, 1, 0, 0])
    box_y_points = y_lower + (y_upper - y_lower) * np.array([0, 0, 1, 1, 0])
    # Get the Iris coordinate sytem of the X coordinate (Y should be the same).
    cs_data1 = x_coord.coord_system
    # Construct an equivalent Cartopy coordinate reference system ("crs").
    crs_data1 = cs_data1.as_cartopy_crs()
    # Draw the rectangle in this crs, with matplotlib "pyplot.plot".
    # NOTE: the 'transform' keyword specifies a non-display coordinate system
    # for the plot points (as used by the "iris.plot" functions).
    plt.plot(
        box_x_points,
        box_y_points,
        transform=crs_data1,
        linewidth=2.0,
        color="white",
        linestyle="--",
    )

    # Mark some particular places with a small circle and a name label...
    # Define some test points with latitude and longitude coordinates.
    city_data = [
        ("London", 51.5072, 0.1275),
        ("Halifax, NS", 44.67, -63.61),
        ("Reykjavik", 64.1333, -21.9333),
    ]
    # Place a single marker point and a text annotation at each place.
    for name, lat, lon in city_data:
        plt.plot(
            lon,
            lat,
            marker="o",
            markersize=7.0,
            markeredgewidth=2.5,
            markerfacecolor="black",
            markeredgecolor="white",
            transform=crs_latlon,
        )
        # NOTE: the "plt.annotate call" does not have a "transform=" keyword,
        # so for this one we transform the coordinates with a Cartopy call.
        at_x, at_y = ax.projection.transform_point(
            lon, lat, src_crs=crs_latlon
        )
        plt.annotate(
            name,
            xy=(at_x, at_y),
            xytext=(30, 20),
            textcoords="offset points",
            color="black",
            backgroundcolor="white",
            size="large",
            arrowprops=dict(arrowstyle="->", color="white", linewidth=2.5),
        )

    # Add a title, and display.
    plt.title(
        "A pseudocolour plot on the {} projection,\n"
        "with overlaid contours.".format(projection_name)
    )
    iplt.show()


def main():
    # Demonstrate with two different display projections.
    make_plot("Equidistant Cylindrical", ccrs.PlateCarree())
    make_plot("North Polar Stereographic", ccrs.NorthPolarStereo())


if __name__ == "__main__":
    main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
A package for converting cubes to and from specific file formats.

"""

from iris.io.format_picker import (
    FileExtension,
    FormatAgent,
    FormatSpecification,
    MagicNumber,
    UriProtocol,
    LeadingLine,
)
from . import abf
from . import um
from . import name
from . import netcdf
from . import nimrod
from . import pp


__all__ = ["FORMAT_AGENT"]


FORMAT_AGENT = FormatAgent()
FORMAT_AGENT.__doc__ = (
    "The FORMAT_AGENT is responsible for identifying the "
    "format of a given URI. New formats can be added "
    "with the **add_spec** method."
)


#
# PP files.
#
FORMAT_AGENT.add_spec(
    FormatSpecification(
        "UM Post Processing file (PP)",
        MagicNumber(4),
        0x00000100,
        pp.load_cubes,
        priority=5,
        constraint_aware_handler=True,
    )
)


FORMAT_AGENT.add_spec(
    FormatSpecification(
        "UM Post Processing file (PP) little-endian",
        MagicNumber(4),
        0x00010000,
        pp.load_cubes_little_endian,
        priority=3,
        constraint_aware_handler=True,
    )
)


#
# GRIB files.
#
def _load_grib(*args, **kwargs):
    try:
        from iris_grib import load_cubes
    except ImportError:
        raise RuntimeError(
            "Unable to load GRIB file - "
            '"iris_grib" package is not installed.'
        )

    return load_cubes(*args, **kwargs)


# NB. Because this is such a "fuzzy" check, we give this a very low
# priority to avoid collateral damage from false positives.
FORMAT_AGENT.add_spec(
    FormatSpecification(
        "GRIB",
        MagicNumber(100),
        lambda header_bytes: b"GRIB" in header_bytes,
        _load_grib,
        priority=1,
    )
)


#
# netCDF files.
#
FORMAT_AGENT.add_spec(
    FormatSpecification(
        "NetCDF", MagicNumber(4), 0x43444601, netcdf.load_cubes, priority=5
    )
)


FORMAT_AGENT.add_spec(
    FormatSpecification(
        "NetCDF 64 bit offset format",
        MagicNumber(4),
        0x43444602,
        netcdf.load_cubes,
        priority=5,
    )
)


# This covers both v4 and v4 classic model.
FORMAT_AGENT.add_spec(
    FormatSpecification(
        "NetCDF_v4",
        MagicNumber(8),
        0x894844460D0A1A0A,
        netcdf.load_cubes,
        priority=5,
    )
)


_nc_dap = FormatSpecification(
    "NetCDF OPeNDAP",
    UriProtocol(),
    lambda protocol: protocol in ["http", "https"],
    netcdf.load_cubes,
    priority=6,
)
FORMAT_AGENT.add_spec(_nc_dap)
del _nc_dap

#
# UM Fieldsfiles.
#
FORMAT_AGENT.add_spec(
    FormatSpecification(
        "UM Fieldsfile (FF) pre v3.1",
        MagicNumber(8),
        0x000000000000000F,
        um.load_cubes,
        priority=3,
        constraint_aware_handler=True,
    )
)


FORMAT_AGENT.add_spec(
    FormatSpecification(
        "UM Fieldsfile (FF) post v5.2",
        MagicNumber(8),
        0x0000000000000014,
        um.load_cubes,
        priority=4,
        constraint_aware_handler=True,
    )
)


FORMAT_AGENT.add_spec(
    FormatSpecification(
        "UM Fieldsfile (FF) ancillary",
        MagicNumber(8),
        0xFFFFFFFFFFFF8000,
        um.load_cubes,
        priority=3,
        constraint_aware_handler=True,
    )
)


FORMAT_AGENT.add_spec(
    FormatSpecification(
        "UM Fieldsfile (FF) converted " "with ieee to 32 bit",
        MagicNumber(4),
        0x00000014,
        um.load_cubes_32bit_ieee,
        priority=3,
        constraint_aware_handler=True,
    )
)


FORMAT_AGENT.add_spec(
    FormatSpecification(
        "UM Fieldsfile (FF) ancillary " "converted with ieee to 32 bit",
        MagicNumber(4),
        0xFFFF8000,
        um.load_cubes_32bit_ieee,
        priority=3,
        constraint_aware_handler=True,
    )
)


#
# NIMROD files.
#
FORMAT_AGENT.add_spec(
    FormatSpecification(
        "NIMROD", MagicNumber(4), 0x00000200, nimrod.load_cubes, priority=3
    )
)

#
# NAME files.
#
FORMAT_AGENT.add_spec(
    FormatSpecification(
        "NAME III",
        LeadingLine(),
        lambda line: line.lstrip().startswith(b"NAME III"),
        name.load_cubes,
        priority=5,
    )
)


#
# ABF/ABL
#
FORMAT_AGENT.add_spec(
    FormatSpecification(
        "ABF", FileExtension(), ".abf", abf.load_cubes, priority=3
    )
)


FORMAT_AGENT.add_spec(
    FormatSpecification(
        "ABL", FileExtension(), ".abl", abf.load_cubes, priority=3
    )
)



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""Unit tests for :func:`iris.fileformats.dot._dot_path`."""

# Import iris.tests first so that some things can be initialised before
# importing anything else.
import iris.tests as tests

import os.path
import subprocess
from unittest import mock

from iris.fileformats.dot import _dot_path, _DOT_EXECUTABLE_PATH


class Test(tests.IrisTest):
    def setUp(self):
        # Because _dot_path is triggered by the initial import we
        # reset the caching status to allow us to see what happens
        # under different circumstances.
        self.patch("iris.fileformats.dot._DOT_CHECKED", new=False)
        # Also patch the private path variable to the existing value (i.e. no
        # change), and restore it after each test:  As these tests modify it,
        # that can potentially break subsequent 'normal' behaviour.
        self.patch(
            "iris.fileformats.dot._DOT_EXECUTABLE_PATH", _DOT_EXECUTABLE_PATH
        )

    def test_valid_absolute_path(self):
        # Override the configuration value for System.dot_path
        real_path = os.path.abspath(__file__)
        assert os.path.exists(real_path) and os.path.isabs(real_path)
        with mock.patch("iris.config.get_option", return_value=real_path):
            result = _dot_path()
        self.assertEqual(result, real_path)

    def test_invalid_absolute_path(self):
        # Override the configuration value for System.dot_path
        dummy_path = "/not_a_real_path" * 10
        assert not os.path.exists(dummy_path)
        with mock.patch("iris.config.get_option", return_value=dummy_path):
            result = _dot_path()
        self.assertIsNone(result)

    def test_valid_relative_path(self):
        # Override the configuration value for System.dot_path
        dummy_path = "not_a_real_path" * 10
        assert not os.path.exists(dummy_path)
        with mock.patch("iris.config.get_option", return_value=dummy_path):
            # Pretend we have a valid installation of dot
            with mock.patch("subprocess.check_output"):
                result = _dot_path()
        self.assertEqual(result, dummy_path)

    def test_valid_relative_path_broken_install(self):
        # Override the configuration value for System.dot_path
        dummy_path = "not_a_real_path" * 10
        assert not os.path.exists(dummy_path)
        with mock.patch("iris.config.get_option", return_value=dummy_path):
            # Pretend we have a broken installation of dot
            error = subprocess.CalledProcessError(-5, "foo", "bar")
            with mock.patch("subprocess.check_output", side_effect=error):
                result = _dot_path()
        self.assertIsNone(result)


if __name__ == "__main__":
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Unit tests for the function :func:`iris.analysis.maths._inplace_common_checks`.

"""

# Import iris.tests first so that some things can be initialised before
# importing anything else.
import iris.tests as tests

from iris.cube import Cube
import numpy as np

from iris.analysis.maths import _inplace_common_checks


class Test(tests.IrisTest):
    # `_inplace_common_checks` is a pass-through function that does not return
    # anything but will fail iff `cube` and `other` have integer dtype. Thus in
    # a sense we only want to test the failing cases. Doing so, however, leaves
    # us open to the case where currently known good cases fail silently.
    # To avoid this all the known good cases are also tested by relying on the
    # fact that functions with no return value implicitly return `None`. If
    # these currently known good cases ever changed these tests would start
    # failing and indicate something was wrong.
    def setUp(self):
        self.scalar_int = 5
        self.scalar_float = 5.5

        self.float_data = np.array([8, 9], dtype=np.float64)
        self.int_data = np.array([9, 8], dtype=np.int64)
        self.uint_data = np.array([9, 8], dtype=np.uint64)

        self.float_cube = Cube(self.float_data)
        self.int_cube = Cube(self.int_data)
        self.uint_cube = Cube(self.uint_data)

        self.op = "addition"
        self.emsg = "Cannot perform inplace {}".format(self.op)

    def test_float_cubes(self):
        result = _inplace_common_checks(
            self.float_cube, self.float_cube, self.op
        )
        self.assertIsNone(result)

    def test_int_cubes(self):
        result = _inplace_common_checks(self.int_cube, self.int_cube, self.op)
        self.assertIsNone(result)

    def test_uint_cubes(self):
        result = _inplace_common_checks(
            self.uint_cube, self.uint_cube, self.op
        )
        self.assertIsNone(result)

    def test_float_cube_int_cube(self):
        result = _inplace_common_checks(
            self.float_cube, self.int_cube, self.op
        )
        self.assertIsNone(result)

    def test_float_cube_uint_cube(self):
        result = _inplace_common_checks(
            self.float_cube, self.uint_cube, self.op
        )
        self.assertIsNone(result)

    def test_int_cube_float_cube(self):
        with self.assertRaisesRegex(ArithmeticError, self.emsg):
            _inplace_common_checks(self.int_cube, self.float_cube, self.op)

    def test_uint_cube_float_cube(self):
        with self.assertRaisesRegex(ArithmeticError, self.emsg):
            _inplace_common_checks(self.uint_cube, self.float_cube, self.op)

    def test_float_cube__scalar_int(self):
        result = _inplace_common_checks(
            self.float_cube, self.scalar_int, self.op
        )
        self.assertIsNone(result)

    def test_float_cube__scalar_float(self):
        result = _inplace_common_checks(
            self.float_cube, self.scalar_float, self.op
        )
        self.assertIsNone(result)

    def test_float_cube__int_array(self):
        result = _inplace_common_checks(
            self.float_cube, self.int_data, self.op
        )
        self.assertIsNone(result)

    def test_float_cube__float_array(self):
        result = _inplace_common_checks(
            self.float_cube, self.float_data, self.op
        )
        self.assertIsNone(result)

    def test_int_cube__scalar_int(self):
        result = _inplace_common_checks(
            self.int_cube, self.scalar_int, self.op
        )
        self.assertIsNone(result)

    def test_int_cube_uint_cube(self):
        result = _inplace_common_checks(self.int_cube, self.uint_cube, self.op)
        self.assertIsNone(result)

    def test_uint_cube_uint_cube(self):
        result = _inplace_common_checks(
            self.uint_cube, self.uint_cube, self.op
        )
        self.assertIsNone(result)

    def test_uint_cube_int_cube(self):
        with self.assertRaisesRegex(ArithmeticError, self.emsg):
            _inplace_common_checks(self.uint_cube, self.int_cube, self.op)

    def test_int_cube__scalar_float(self):
        with self.assertRaisesRegex(ArithmeticError, self.emsg):
            _inplace_common_checks(self.int_cube, self.scalar_float, self.op)

    def test_int_cube__int_array(self):
        result = _inplace_common_checks(self.int_cube, self.int_cube, self.op)
        self.assertIsNone(result)

    def test_int_cube__float_array(self):
        with self.assertRaisesRegex(ArithmeticError, self.emsg):
            _inplace_common_checks(self.int_cube, self.float_data, self.op)

    def test_uint_cube__scalar_float(self):
        with self.assertRaisesRegex(ArithmeticError, self.emsg):
            _inplace_common_checks(self.uint_cube, self.scalar_float, self.op)

    def test_uint_cube__int_array(self):
        with self.assertRaisesRegex(ArithmeticError, self.emsg):
            _inplace_common_checks(self.uint_cube, self.int_cube, self.op)

    def test_uint_cube__float_array(self):
        with self.assertRaisesRegex(ArithmeticError, self.emsg):
            _inplace_common_checks(self.uint_cube, self.float_data, self.op)


if __name__ == "__main__":
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.

import warnings

import iris
from iris.fileformats._ff_cross_references import STASH_TRANS
from iris.aux_factory import HybridHeightFactory, HybridPressureFactory
from iris.fileformats.um_cf_map import CF_TO_LBFC
from iris.fileformats._pp_lbproc_pairs import LBPROC_MAP
from iris.fileformats.rules import (
    aux_factory,
    has_aux_factory,
    scalar_cell_method,
    scalar_coord,
    vector_coord,
)
from iris.util import is_regular, regular_step
import cftime


def _basic_coord_system_rules(cube, pp):
    """
    Rules for setting the coord system of the PP field.

    Args:
        cube: the cube being saved as a series of PP fields.
        pp: the current PP field having save rules applied.

    Returns:
        The PP field with updated metadata.

    """
    if (
        cube.coord_system("GeogCS") is not None
        or cube.coord_system(None) is None
    ):
        pp.bplat = 90
        pp.bplon = 0
    elif cube.coord_system("RotatedGeogCS") is not None:
        pp.bplat = cube.coord_system("RotatedGeogCS").grid_north_pole_latitude
        pp.bplon = cube.coord_system("RotatedGeogCS").grid_north_pole_longitude
    return pp


def _um_version_rules(cube, pp):
    from_um_str = "Data from Met Office Unified Model"
    source_attr = cube.attributes.get("source")
    if source_attr is not None:
        um_version = source_attr.rsplit(from_um_str, 1)

    if (
        "um_version" not in cube.attributes
        and "source" in cube.attributes
        and len(um_version) > 1
        and len(um_version[1]) == 0
    ):
        # UM - no version number.
        pp.lbsrce = 1111
    elif (
        "um_version" not in cube.attributes
        and "source" in cube.attributes
        and len(um_version) > 1
        and len(um_version[1]) > 0
    ):
        # UM - with version number.
        pp.lbsrce = int(float(um_version[1]) * 1000000) + 1111
    elif "um_version" in cube.attributes:
        # UM - from 'um_version' attribute.
        um_ver_minor = int(cube.attributes["um_version"].split(".")[1])
        um_ver_major = int(cube.attributes["um_version"].split(".")[0])
        pp.lbsrce = 1111 + 10000 * um_ver_minor + 1000000 * um_ver_major
    return pp


def _stash_rules(cube, pp):
    """
    Attributes rules for setting the STASH attribute of the PP field.

    Args:
        cube: the cube being saved as a series of PP fields.
        pp: the current PP field having save rules applied.

    Returns:
        The PP field with updated metadata.

    """
    if "STASH" in cube.attributes:
        stash = cube.attributes["STASH"]
        if isinstance(stash, iris.fileformats.pp.STASH):
            pp.lbuser[3] = 1000 * (stash.section or 0) + (stash.item or 0)
            pp.lbuser[6] = stash.model or 0
    return pp


def _general_time_rules(cube, pp):
    """
    Rules for setting time metadata of the PP field.

    Args:
        cube: the cube being saved as a series of PP fields.
        pp: the current PP field having save rules applied.

    Returns:
        The PP field with updated metadata.

    """
    time_coord = scalar_coord(cube, "time")
    fp_coord = scalar_coord(cube, "forecast_period")
    frt_coord = scalar_coord(cube, "forecast_reference_time")
    clim_season_coord = scalar_coord(cube, "clim_season")

    cm_time_mean = scalar_cell_method(cube, "mean", "time")
    cm_time_min = scalar_cell_method(cube, "minimum", "time")
    cm_time_max = scalar_cell_method(cube, "maximum", "time")

    # No forecast.
    if time_coord is not None and fp_coord is None and frt_coord is None:
        pp.lbtim.ia = 0
        pp.lbtim.ib = 0
        pp.t1 = time_coord.units.num2date(time_coord.points[0])
        pp.t2 = cftime.datetime(0, 0, 0)

    # Forecast.
    if (
        time_coord is not None
        and not time_coord.has_bounds()
        and fp_coord is not None
    ):
        pp.lbtim.ia = 0
        pp.lbtim.ib = 1
        pp.t1 = time_coord.units.num2date(time_coord.points[0])
        pp.t2 = time_coord.units.num2date(
            time_coord.points[0] - fp_coord.points[0]
        )
        pp.lbft = fp_coord.points[0]

    # Time mean (non-climatological).
    # XXX This only works when we have a single timestep.
    if (
        time_coord is not None
        and time_coord.has_bounds()
        and clim_season_coord is None
        and fp_coord is not None
        and fp_coord.has_bounds()
    ):
        # XXX How do we know *which* time to use if there are more than
        # one? *Can* there be more than one?
        pp.lbtim.ib = 2
        pp.t1 = time_coord.units.num2date(time_coord.bounds[0, 0])
        pp.t2 = time_coord.units.num2date(time_coord.bounds[0, 1])
        pp.lbft = fp_coord.units.convert(fp_coord.bounds[0, 1], "hours")

    if (
        time_coord is not None
        and time_coord.has_bounds()
        and clim_season_coord is None
        and fp_coord is None
        and frt_coord is not None
    ):
        # Handle missing forecast period using time and forecast ref time.
        pp.lbtim.ib = 2
        pp.t1 = time_coord.units.num2date(time_coord.bounds[0, 0])
        pp.t2 = time_coord.units.num2date(time_coord.bounds[0, 1])
        stop = time_coord.units.convert(
            time_coord.bounds[0, 1], "hours since epoch"
        )
        start = frt_coord.units.convert(
            frt_coord.points[0], "hours since epoch"
        )
        pp.lbft = stop - start

    if (
        time_coord is not None
        and time_coord.has_bounds()
        and clim_season_coord is None
        and cm_time_mean is not None
    ):
        pp.lbtim.ib = 2
        pp.t1 = time_coord.units.num2date(time_coord.bounds[0, 0])
        pp.t2 = time_coord.units.num2date(time_coord.bounds[0, 1])

    if (
        time_coord is not None
        and time_coord.has_bounds()
        and clim_season_coord is None
        and cm_time_mean is not None
        and cm_time_mean.intervals != ()
        and cm_time_mean.intervals[0].endswith("hour")
    ):
        pp.lbtim.ia = int(cm_time_mean.intervals[0][:-5])

    if (
        time_coord is not None
        and time_coord.has_bounds()
        and clim_season_coord is None
        and (fp_coord is not None or frt_coord is not None)
        and (
            cm_time_mean is None
            or cm_time_mean.intervals == ()
            or not cm_time_mean.intervals[0].endswith("hour")
        )
    ):
        pp.lbtim.ia = 0

    # If the cell methods contain a minimum then overwrite lbtim.ia with this
    # interval.
    if (
        time_coord is not None
        and time_coord.has_bounds()
        and clim_season_coord is None
        and (fp_coord is not None or frt_coord is not None)
        and cm_time_min is not None
        and cm_time_min.intervals != ()
        and cm_time_min.intervals[0].endswith("hour")
    ):
        # Set lbtim.ia with the integer part of the cell method's interval
        # e.g. if interval is '24 hour' then lbtim.ia becomes 24.
        pp.lbtim.ia = int(cm_time_min.intervals[0][:-5])

    # If the cell methods contain a maximum then overwrite lbtim.ia with this
    # interval.
    if (
        time_coord is not None
        and time_coord.has_bounds()
        and clim_season_coord is None
        and (fp_coord is not None or frt_coord is not None)
        and cm_time_max is not None
        and cm_time_max.intervals != ()
        and cm_time_max.intervals[0].endswith("hour")
    ):
        # Set lbtim.ia with the integer part of the cell method's interval
        # e.g. if interval is '1 hour' then lbtim.ia becomes 1.
        pp.lbtim.ia = int(cm_time_max.intervals[0][:-5])

    if time_coord is not None and time_coord.has_bounds():
        lower_bound_yr = time_coord.units.num2date(
            time_coord.bounds[0, 0]
        ).year
        upper_bound_yr = time_coord.units.num2date(
            time_coord.bounds[0, 1]
        ).year
    else:
        lower_bound_yr = None
        upper_bound_yr = None

    # Climatological time means.
    if (
        time_coord is not None
        and time_coord.has_bounds()
        and lower_bound_yr == upper_bound_yr
        and fp_coord is not None
        and fp_coord.has_bounds()
        and clim_season_coord is not None
        and "clim_season" in cube.cell_methods[-1].coord_names
    ):
        # Climatological time mean - single year.
        pp.lbtim.ia = 0
        pp.lbtim.ib = 2
        pp.t1 = time_coord.units.num2date(time_coord.bounds[0, 0])
        pp.t2 = time_coord.units.num2date(time_coord.bounds[0, 1])
        pp.lbft = fp_coord.units.convert(fp_coord.bounds[0, 1], "hours")

    elif (
        time_coord is not None
        and time_coord.has_bounds()
        and lower_bound_yr != upper_bound_yr
        and fp_coord is not None
        and fp_coord.has_bounds()
        and clim_season_coord is not None
        and "clim_season" in cube.cell_methods[-1].coord_names
        and clim_season_coord.points[0] == "djf"
    ):
        # Climatological time mean - spanning years - djf.
        pp.lbtim.ia = 0
        pp.lbtim.ib = 3
        pp.t1 = time_coord.units.num2date(time_coord.bounds[0, 0])
        pp.t2 = time_coord.units.num2date(time_coord.bounds[0, 1])
        if pp.t1.month == 12:
            pp.t1 = cftime.datetime(pp.t1.year)
        else:
            pp.t1 = cftime.datetime(pp.t1.year - 1, 12, 1, 0, 0, 0)
        pp.t2 = cftime.datetime(pp.t2.year, 3, 1, 0, 0, 0)
        _conditional_warning(
            time_coord.bounds[0, 0] != time_coord.units.date2num(pp.t1),
            "modified t1 for climatological seasonal mean",
        )
        _conditional_warning(
            time_coord.bounds[0, 1] != time_coord.units.date2num(pp.t2),
            "modified t2 for climatological seasonal mean",
        )
        pp.lbft = fp_coord.units.convert(fp_coord.bounds[0, 1], "hours")

    elif (
        time_coord is not None
        and time_coord.has_bounds()
        and lower_bound_yr != upper_bound_yr
        and fp_coord is not None
        and fp_coord.has_bounds()
        and clim_season_coord is not None
        and "clim_season" in cube.cell_methods[-1].coord_names
        and clim_season_coord.points[0] == "mam"
    ):
        # Climatological time mean - spanning years - mam.
        pp.lbtim.ia = 0
        pp.lbtim.ib = 3
        # TODO: wut?
        pp.t1 = time_coord.units.num2date(time_coord.bounds[0, 0])
        pp.t2 = time_coord.units.num2date(time_coord.bounds[0, 1])
        pp.t1 = cftime.datetime(pp.t1.year, 3, 1, 0, 0, 0)
        pp.t2 = cftime.datetime(pp.t2.year, 6, 1, 0, 0, 0)
        _conditional_warning(
            time_coord.bounds[0, 0] != time_coord.units.date2num(pp.t1),
            "modified t1 for climatological seasonal mean",
        )
        _conditional_warning(
            time_coord.bounds[0, 1] != time_coord.units.date2num(pp.t2),
            "modified t2 for climatological seasonal mean",
        )
        pp.lbft = fp_coord.units.convert(fp_coord.bounds[0, 1], "hours")

    elif (
        time_coord is not None
        and time_coord.has_bounds()
        and lower_bound_yr != upper_bound_yr
        and fp_coord is not None
        and fp_coord.has_bounds()
        and clim_season_coord is not None
        and "clim_season" in cube.cell_methods[-1].coord_names
        and clim_season_coord.points[0] == "jja"
    ):
        # Climatological time mean - spanning years - jja.
        pp.lbtim.ia = 0
        pp.lbtim.ib = 3
        # TODO: wut?
        pp.t1 = time_coord.units.num2date(time_coord.bounds[0, 0])
        pp.t2 = time_coord.units.num2date(time_coord.bounds[0, 1])
        pp.t1 = cftime.datetime(pp.t1.year, 6, 1, 0, 0, 0)
        pp.t2 = cftime.datetime(pp.t2.year, 9, 1, 0, 0, 0)
        _conditional_warning(
            time_coord.bounds[0, 0] != time_coord.units.date2num(pp.t1),
            "modified t1 for climatological seasonal mean",
        )
        _conditional_warning(
            time_coord.bounds[0, 1] != time_coord.units.date2num(pp.t2),
            "modified t2 for climatological seasonal mean",
        )
        pp.lbft = fp_coord.units.convert(fp_coord.bounds[0, 1], "hours")

    elif (
        time_coord is not None
        and time_coord.has_bounds()
        and lower_bound_yr != upper_bound_yr
        and fp_coord is not None
        and fp_coord.has_bounds()
        and clim_season_coord is not None
        and "clim_season" in cube.cell_methods[-1].coord_names
        and clim_season_coord.points[0] == "son"
    ):
        # Climatological time mean - spanning years - son.
        pp.lbtim.ia = 0
        pp.lbtim.ib = 3
        # TODO: wut?
        pp.t1 = time_coord.units.num2date(time_coord.bounds[0, 0])
        pp.t2 = time_coord.units.num2date(time_coord.bounds[0, 1])
        pp.t1 = cftime.datetime(pp.t1.year, 9, 1, 0, 0, 0)
        pp.t2 = cftime.datetime(pp.t2.year, 12, 1, 0, 0, 0)
        _conditional_warning(
            time_coord.bounds[0, 0] != time_coord.units.date2num(pp.t1),
            "modified t1 for climatological seasonal mean",
        )
        _conditional_warning(
            time_coord.bounds[0, 1] != time_coord.units.date2num(pp.t2),
            "modified t2 for climatological seasonal mean",
        )
        pp.lbft = fp_coord.units.convert(fp_coord.bounds[0, 1], "hours")

    return pp


def _calendar_rules(cube, pp):
    """
    Rules for setting the calendar of the PP field.

    Args:
        cube: the cube being saved as a series of PP fields.
        pp: the current PP field having save rules applied.

    Returns:
        The PP field with updated metadata.

    """
    time_coord = scalar_coord(cube, "time")
    if time_coord is not None:
        if time_coord.units.calendar == "360_day":
            pp.lbtim.ic = 2
        elif time_coord.units.calendar == "gregorian":
            pp.lbtim.ic = 1
        elif time_coord.units.calendar == "365_day":
            pp.lbtim.ic = 4
    return pp


def _grid_and_pole_rules(cube, pp):
    """
    Rules for setting the horizontal grid and pole location of the PP field.

    Args:
        cube: the cube being saved as a series of PP fields.
        pp: the current PP field having save rules applied.

    Returns:
        The PP field with updated metadata.

    """
    lon_coord = vector_coord(cube, "longitude")
    grid_lon_coord = vector_coord(cube, "grid_longitude")
    lat_coord = vector_coord(cube, "latitude")
    grid_lat_coord = vector_coord(cube, "grid_latitude")

    if lon_coord and not is_regular(lon_coord):
        pp.bzx = 0
        pp.bdx = 0
        pp.lbnpt = lon_coord.shape[0]
        pp.x = lon_coord.points
    elif grid_lon_coord and not is_regular(grid_lon_coord):
        pp.bzx = 0
        pp.bdx = 0
        pp.lbnpt = grid_lon_coord.shape[0]
        pp.x = grid_lon_coord.points
    elif lon_coord and is_regular(lon_coord):
        pp.bzx = lon_coord.points[0] - regular_step(lon_coord)
        pp.bdx = regular_step(lon_coord)
        pp.lbnpt = len(lon_coord.points)
    elif grid_lon_coord and is_regular(grid_lon_coord):
        pp.bzx = grid_lon_coord.points[0] - regular_step(grid_lon_coord)
        pp.bdx = regular_step(grid_lon_coord)
        pp.lbnpt = len(grid_lon_coord.points)

    if lat_coord and not is_regular(lat_coord):
        pp.bzy = 0
        pp.bdy = 0
        pp.lbrow = lat_coord.shape[0]
        pp.y = lat_coord.points
    elif grid_lat_coord and not is_regular(grid_lat_coord):
        pp.bzy = 0
        pp.bdy = 0
        pp.lbrow = grid_lat_coord.shape[0]
        pp.y = grid_lat_coord.points
    elif lat_coord and is_regular(lat_coord):
        pp.bzy = lat_coord.points[0] - regular_step(lat_coord)
        pp.bdy = regular_step(lat_coord)
        pp.lbrow = len(lat_coord.points)
    elif grid_lat_coord and is_regular(grid_lat_coord):
        pp.bzy = grid_lat_coord.points[0] - regular_step(grid_lat_coord)
        pp.bdy = regular_step(grid_lat_coord)
        pp.lbrow = len(grid_lat_coord.points)

    # Check if we have a rotated coord system.
    if cube.coord_system("RotatedGeogCS") is not None:
        pp.lbcode = int(pp.lbcode) + 100

    # Check if we have a circular x-coord.
    for lon_coord in (lon_coord, grid_lon_coord):
        if lon_coord is not None:
            if lon_coord.circular:
                pp.lbhem = 0
            else:
                pp.lbhem = 3

    return pp


def _non_std_cross_section_rules(cube, pp):
    """
    Rules for applying non-standard cross-sections to the PP field.

    Args:
        cube: the cube being saved as a series of PP fields.
        pp: the current PP field having save rules applied.

    Returns:
        The PP field with updated metadata.

    """
    # Define commonly-used coords.
    air_pres_coord = vector_coord(cube, "air_pressure")
    depth_coord = vector_coord(cube, "depth")
    eta_coord = vector_coord(cube, "eta")
    lat_coord = vector_coord(cube, "latitude")
    time_coord = vector_coord(cube, "time")

    # Non-standard cross-section with bounds - x=latitude, y=air_pressure.
    if (
        air_pres_coord is not None
        and not air_pres_coord.circular
        and air_pres_coord.has_bounds()
        and lat_coord is not None
        and not lat_coord.circular
        and lat_coord.has_bounds()
    ):
        pp.lbcode = 10000 + int(100 * 10) + 1
        pp.bgor = 0
        pp.y = air_pres_coord.points
        pp.y_lower_bound = air_pres_coord.bounds[:, 0]
        pp.y_upper_bound = air_pres_coord.bounds[:, 1]
        pp.x = lat_coord.points
        pp.x_lower_bound = lat_coord.bounds[:, 0]
        pp.x_upper_bound = lat_coord.bounds[:, 1]
        pp.lbrow = air_pres_coord.shape[0]
        pp.lbnpt = lat_coord.shape[0]
        pp.bzx = pp.bzy = pp.bdx = pp.bdy = 0

    # Non-standard cross-section with bounds - x=latitude, y=depth.
    if (
        depth_coord is not None
        and not depth_coord.circular
        and depth_coord.has_bounds()
        and lat_coord is not None
        and not lat_coord.circular
        and lat_coord.has_bounds()
    ):
        pp.lbcode = 10000 + int(100 * 10) + 4
        pp.bgor = 0
        pp.y = depth_coord.points
        pp.y_lower_bound = depth_coord.bounds[:, 0]
        pp.y_upper_bound = depth_coord.bounds[:, 1]
        pp.x = lat_coord.points
        pp.x_lower_bound = lat_coord.bounds[:, 0]
        pp.x_upper_bound = lat_coord.bounds[:, 1]
        pp.lbrow = depth_coord.shape[0]
        pp.lbnpt = lat_coord.shape[0]
        pp.bzx = pp.bzy = pp.bdx = pp.bdy = 0

    # Non-standard cross-section with bounds - x=latitude, y=eta.
    if (
        eta_coord is not None
        and not eta_coord.circular
        and eta_coord.has_bounds()
        and lat_coord is not None
        and not lat_coord.circular
        and lat_coord.has_bounds()
    ):
        pp.lbcode = 10000 + int(100 * 10) + 3
        pp.bgor = 0
        pp.y = eta_coord.points
        pp.y_lower_bound = eta_coord.bounds[:, 0]
        pp.y_upper_bound = eta_coord.bounds[:, 1]
        pp.x = lat_coord.points
        pp.x_lower_bound = lat_coord.bounds[:, 0]
        pp.x_upper_bound = lat_coord.bounds[:, 1]
        pp.lbrow = eta_coord.shape[0]
        pp.lbnpt = lat_coord.shape[0]
        pp.bzx = pp.bzy = pp.bdx = pp.bdy = 0

    # Non-standard cross-section with bounds - x=days (360 calendar), y=depth.
    if (
        depth_coord is not None
        and not depth_coord.circular
        and depth_coord.has_bounds()
        and time_coord is not None
        and not time_coord.circular
        and time_coord.has_bounds()
    ):
        pp.lbcode = 10000 + int(100 * 23) + 4
        pp.bgor = 0
        pp.y = depth_coord.points
        pp.y_lower_bound = depth_coord.bounds[:, 0]
        pp.y_upper_bound = depth_coord.bounds[:, 1]
        pp.x = time_coord.points
        pp.x_lower_bound = time_coord.bounds[:, 0]
        pp.x_upper_bound = time_coord.bounds[:, 1]
        pp.lbrow = depth_coord.shape[0]
        pp.lbnpt = time_coord.shape[0]
        pp.bzx = pp.bzy = pp.bdx = pp.bdy = 0

    # Non-standard cross-section with bounds -
    # x=days (360 calendar), y=air_pressure.
    if (
        air_pres_coord is not None
        and not air_pres_coord.circular
        and air_pres_coord.has_bounds()
        and time_coord is not None
        and not time_coord.circular
        and time_coord.has_bounds()
    ):
        pp.lbcode = 10000 + int(100 * 23) + 1
        pp.bgor = 0
        pp.y = air_pres_coord.points
        pp.y_lower_bound = air_pres_coord.bounds[:, 0]
        pp.y_upper_bound = air_pres_coord.bounds[:, 1]
        pp.x = time_coord.points
        pp.x_lower_bound = time_coord.bounds[:, 0]
        pp.x_upper_bound = time_coord.bounds[:, 1]
        pp.lbrow = air_pres_coord.shape[0]
        pp.lbnpt = time_coord.shape[0]
        pp.bzx = pp.bzy = pp.bdx = pp.bdy = 0

    return pp


def _lbproc_rules(cube, pp):
    """
    Rules for setting the horizontal grid and pole location of the PP field.

    Note: `pp.lbproc` must be set to 0 before these rules are run.

    Args:
        cube: the cube being saved as a series of PP fields.
        pp: the current PP field having save rules applied.

    Returns:
        The PP field with updated metadata.

    """
    # Basic setting (this may be overridden by subsequent rules).
    pp.lbproc = 0

    if cube.attributes.get("ukmo__process_flags", None):
        pp.lbproc += sum(
            [
                LBPROC_MAP[name]
                for name in cube.attributes["ukmo__process_flags"]
            ]
        )

    # Zonal-mean: look for a CellMethod which is a "mean" over "longitude" or
    # "grid_longitude".
    if (
        scalar_cell_method(cube, "mean", "longitude") is not None
        or scalar_cell_method(cube, "mean", "grid_longitude") is not None
    ):
        pp.lbproc += 64

    # Time-mean: look for a CellMethod which is a "mean" over "time".
    if scalar_cell_method(cube, "mean", "time") is not None:
        pp.lbproc += 128

    # Time-minimum: look for a CellMethod which is a "minimum" over "time".
    if scalar_cell_method(cube, "minimum", "time") is not None:
        pp.lbproc += 4096

    # Time-maximum: look for a CellMethod which is a "maximum" over "time".
    if scalar_cell_method(cube, "maximum", "time") is not None:
        pp.lbproc += 8192

    return pp


def _vertical_rules(cube, pp):
    """
    Rules for setting vertical levels for the PP field.

    Args:
        cube: the cube being saved as a series of PP fields.
        pp: the current PP field having save rules applied.

    Returns:
        The PP field with updated metadata.

    """
    # Define commonly-used coords.
    air_pres_coord = scalar_coord(cube, "air_pressure")
    apt_coord = scalar_coord(cube, "air_potential_temperature")
    depth_coord = scalar_coord(cube, "depth")
    height_coord = scalar_coord(cube, "height")
    level_height_coord = scalar_coord(cube, "level_height")
    mln_coord = scalar_coord(cube, "model_level_number")
    pressure_coord = scalar_coord(cube, "pressure")
    pseudo_level_coord = scalar_coord(cube, "pseudo_level")
    sigma_coord = scalar_coord(cube, "sigma")
    soil_mln_coord = scalar_coord(cube, "soil_model_level_number")

    # Define commonly-used aux factories.
    try:
        height_factory = aux_factory(cube, HybridHeightFactory)
    except ValueError:
        height_factory = None
    try:
        pressure_factory = aux_factory(cube, HybridPressureFactory)
    except ValueError:
        pressure_factory = None

    # Set `lbuser[5]`.
    if pseudo_level_coord is not None and not pseudo_level_coord.bounds:
        pp.lbuser[4] = pseudo_level_coord.points[0]

    # Single height level.
    if (
        height_coord is not None
        and not height_coord.bounds
        and height_coord.points[0] == 1.5
        and cube.name() == "air_temperature"
    ):
        pp.lbvc = 129
        pp.blev = -1

    if pp.lbvc == 0 and height_coord is not None and not height_coord.bounds:
        pp.lbvc = 1
        pp.blev = cube.coord("height").points[0]

    # Single air_pressure level.
    if air_pres_coord is not None and not air_pres_coord.bounds:
        pp.lbvc = 8
        pp.blev = air_pres_coord.points[0]

    # Single pressure level.
    if pressure_coord is not None and not pressure_coord.bounds:
        pp.lbvc = 8
        pp.blev = pressure_coord.points[0]

    # Single depth level (non cross-section).
    if (
        mln_coord is not None
        and not mln_coord.bounds
        and depth_coord is not None
        and not depth_coord.bounds
    ):
        pp.lbvc = 2
        pp.lblev = mln_coord.points[0]
        pp.blev = depth_coord.points[0]

    # Single depth level (Non-dimensional soil model level).
    if (
        soil_mln_coord is not None
        and not soil_mln_coord.has_bounds()
        and air_pres_coord is None
        and depth_coord is None
        and height_coord is None
        and pressure_coord is None
        and cube.standard_name is not None
        and "soil" in cube.standard_name
    ):
        pp.lbvc = 6
        pp.lblev = soil_mln_coord.points[0]
        pp.blev = pp.lblev
        pp.brsvd[0] = 0
        pp.brlev = 0

    # Single depth level (soil depth).
    if (
        depth_coord is not None
        and depth_coord.has_bounds()
        and air_pres_coord is None
        and soil_mln_coord is None
        and mln_coord is None
        and height_coord is None
        and pressure_coord is None
        and cube.standard_name is not None
        and "soil" in cube.standard_name
    ):
        pp.lbvc = 6
        pp.blev = depth_coord.points[0]
        pp.brsvd[0] = depth_coord.bounds[0, 0]
        pp.brlev = depth_coord.bounds[0, 1]

    # Single potential-temperature level.
    if (
        apt_coord is not None
        and not apt_coord.bounds
        and air_pres_coord is None
        and depth_coord is None
        and height_coord is None
        and pressure_coord is None
        and mln_coord is None
    ):
        pp.lbvc = 19
        pp.lblev = apt_coord.points[0]
        pp.blev = apt_coord.points[0]

    # Single hybrid_height level
    # (without aux factory e.g. due to missing orography).
    if (
        not has_aux_factory(cube, HybridHeightFactory)
        and mln_coord is not None
        and mln_coord.bounds is None
        and level_height_coord is not None
        and level_height_coord.bounds is not None
        and sigma_coord is not None
        and sigma_coord.bounds is not None
    ):
        pp.lbvc = 65
        pp.lblev = mln_coord.points[0]
        pp.blev = level_height_coord.points[0]
        pp.brlev = level_height_coord.bounds[0, 0]
        pp.brsvd[0] = level_height_coord.bounds[0, 1]
        pp.bhlev = sigma_coord.points[0]
        pp.bhrlev = sigma_coord.bounds[0, 0]
        pp.brsvd[1] = sigma_coord.bounds[0, 1]

    # Single hybrid_height level (with aux factory).
    if (
        has_aux_factory(cube, HybridHeightFactory)
        and mln_coord is not None
        and mln_coord.bounds is None
        and height_factory.dependencies["delta"] is not None
        and height_factory.dependencies["delta"].bounds is not None
        and height_factory.dependencies["sigma"] is not None
        and height_factory.dependencies["sigma"].bounds is not None
    ):
        pp.lbvc = 65
        pp.lblev = mln_coord.points[0]
        pp.blev = height_factory.dependencies["delta"].points[0]
        pp.brlev = height_factory.dependencies["delta"].bounds[0, 0]
        pp.brsvd[0] = height_factory.dependencies["delta"].bounds[0, 1]
        pp.bhlev = height_factory.dependencies["sigma"].points[0]
        pp.bhrlev = height_factory.dependencies["sigma"].bounds[0, 0]
        pp.brsvd[1] = height_factory.dependencies["sigma"].bounds[0, 1]

    # Single hybrid pressure level.
    if (
        has_aux_factory(cube, HybridPressureFactory)
        and mln_coord is not None
        and mln_coord.bounds is None
        and pressure_factory.dependencies["delta"] is not None
        and pressure_factory.dependencies["delta"].bounds is not None
        and pressure_factory.dependencies["sigma"] is not None
        and pressure_factory.dependencies["sigma"].bounds is not None
    ):
        pp.lbvc = 9
        pp.lblev = mln_coord.points[0]
        pp.blev = pressure_factory.dependencies["sigma"].points[0]
        pp.brlev = pressure_factory.dependencies["sigma"].bounds[0, 0]
        pp.brsvd[0] = pressure_factory.dependencies["sigma"].bounds[0, 1]
        pp.bhlev = pressure_factory.dependencies["delta"].points[0]
        pp.bhrlev = pressure_factory.dependencies["delta"].bounds[0, 0]
        pp.brsvd[1] = pressure_factory.dependencies["delta"].bounds[0, 1]

    return pp


def _all_other_rules(cube, pp):
    """
    Rules for setting the horizontal grid and pole location of the PP field.

    Args:
        cube: the cube being saved as a series of PP fields.
        pp: the current PP field having save rules applied.

    Returns:
        The PP field with updated metadata.

    """
    # "CFNAME mega-rule."
    check_items = (cube.standard_name, cube.long_name, str(cube.units))
    if check_items in CF_TO_LBFC:
        pp.lbfc = CF_TO_LBFC[check_items]

    # Set STASH code.
    if (
        "STASH" in cube.attributes
        and str(cube.attributes["STASH"]) in STASH_TRANS
    ):
        pp.lbfc = STASH_TRANS[str(cube.attributes["STASH"])].field_code

    return pp


def verify(cube, field):
    # Rules functions.
    field = _basic_coord_system_rules(cube, field)
    field = _um_version_rules(cube, field)
    field = _stash_rules(cube, field)
    field = _general_time_rules(cube, field)
    field = _calendar_rules(cube, field)
    field = _grid_and_pole_rules(cube, field)
    field = _non_std_cross_section_rules(cube, field)
    field = _lbproc_rules(cube, field)
    field = _vertical_rules(cube, field)
    field = _all_other_rules(cube, field)

    return field


# Helper functions used when running the rules.


def _conditional_warning(condition, warning):
    if condition:
        warnings.warn(warning)



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Unit tests for the `iris.fileformats.nimrod_load_rules.vertical_coord`
function.

"""

# Import iris.tests first so that some things can be initialised before
# importing anything else.
import iris.tests as tests

from unittest import mock

from iris.fileformats.nimrod_load_rules import (
    vertical_coord,
    NIMROD_DEFAULT,
    TranslationWarning,
)
from iris.fileformats.nimrod import NimrodField


class Test(tests.IrisTest):
    NIMROD_LOCATION = "iris.fileformats.nimrod_load_rules"

    def setUp(self):
        self.field = mock.Mock(
            vertical_coord=NIMROD_DEFAULT,
            vertical_coord_type=NIMROD_DEFAULT,
            reference_vertical_coord=NIMROD_DEFAULT,
            reference_vertical_coord_type=NIMROD_DEFAULT,
            int_mdi=-32767,
            float32_mdi=NIMROD_DEFAULT,
            spec=NimrodField,
        )
        self.cube = mock.Mock()

    def _call_vertical_coord(
        self,
        vertical_coord_val=None,
        vertical_coord_type=None,
        reference_vertical_coord=None,
        reference_vertical_coord_type=None,
    ):
        if vertical_coord_val:
            self.field.vertical_coord = vertical_coord_val
        if vertical_coord_type:
            self.field.vertical_coord_type = vertical_coord_type
        if reference_vertical_coord:
            self.field.reference_vertical_coord = reference_vertical_coord
        if reference_vertical_coord_type:
            self.field.reference_vertical_coord_type = (
                reference_vertical_coord_type
            )
        vertical_coord(self.cube, self.field)

    def test_unhandled(self):
        with mock.patch("warnings.warn") as warn:
            self._call_vertical_coord(
                vertical_coord_val=1.0, vertical_coord_type=-1
            )
        warn.assert_called_once_with(
            "Vertical coord -1 not yet handled", TranslationWarning
        )

    def test_null(self):
        with mock.patch("warnings.warn") as warn:
            self._call_vertical_coord(vertical_coord_type=NIMROD_DEFAULT)
            self._call_vertical_coord(vertical_coord_type=self.field.int_mdi)
        self.assertEqual(warn.call_count, 0)

    def test_ground_level(self):
        with mock.patch("warnings.warn") as warn:
            self._call_vertical_coord(
                vertical_coord_val=9999.0, vertical_coord_type=0
            )
        self.assertEqual(warn.call_count, 0)


if __name__ == "__main__":
    tests.main()



"""
Global Average Annual Temperature Plot
======================================

Produces a time-series plot of North American temperature forecasts for 2
different emission scenarios. Constraining data to a limited spatial area also
features in this example.

The data used comes from the HadGEM2-AO model simulations for the A1B and E1
scenarios, both of which were derived using the IMAGE Integrated Assessment
Model (Johns et al. 2011; Lowe et al. 2009).

References
----------

   Johns T.C., et al. (2011) Climate change under aggressive mitigation: the
   ENSEMBLES multi-model experiment. Climate Dynamics, Vol 37, No. 9-10,
   doi:10.1007/s00382-011-1005-5.

   Lowe J.A., C.D. Hewitt, D.P. Van Vuuren, T.C. Johns, E. Stehfest, J-F.
   Royer, and P. van der Linden, 2009. New Study For Climate Modeling,
   Analyses, and Scenarios. Eos Trans. AGU, Vol 90, No. 21,
   doi:10.1029/2009EO210001.

.. seealso::

    Further details on the aggregation functionality being used in this example
    can be found in :ref:`cube-statistics`.

"""

import matplotlib.pyplot as plt
import numpy as np

import iris
import iris.analysis.cartography
import iris.plot as iplt
import iris.quickplot as qplt


def main():
    # Load data into three Cubes, one for each set of NetCDF files.
    e1 = iris.load_cube(iris.sample_data_path("E1_north_america.nc"))

    a1b = iris.load_cube(iris.sample_data_path("A1B_north_america.nc"))

    # load in the global pre-industrial mean temperature, and limit the domain
    # to the same North American region that e1 and a1b are at.
    north_america = iris.Constraint(
        longitude=lambda v: 225 <= v <= 315, latitude=lambda v: 15 <= v <= 60
    )
    pre_industrial = iris.load_cube(
        iris.sample_data_path("pre-industrial.pp"), north_america
    )

    # Generate area-weights array. As e1 and a1b are on the same grid we can
    # do this just once and re-use. This method requires bounds on lat/lon
    # coords, so let's add some in sensible locations using the "guess_bounds"
    # method.
    e1.coord("latitude").guess_bounds()
    e1.coord("longitude").guess_bounds()
    e1_grid_areas = iris.analysis.cartography.area_weights(e1)
    pre_industrial.coord("latitude").guess_bounds()
    pre_industrial.coord("longitude").guess_bounds()
    pre_grid_areas = iris.analysis.cartography.area_weights(pre_industrial)

    # Perform the area-weighted mean for each of the datasets using the
    # computed grid-box areas.
    pre_industrial_mean = pre_industrial.collapsed(
        ["latitude", "longitude"], iris.analysis.MEAN, weights=pre_grid_areas
    )
    e1_mean = e1.collapsed(
        ["latitude", "longitude"], iris.analysis.MEAN, weights=e1_grid_areas
    )
    a1b_mean = a1b.collapsed(
        ["latitude", "longitude"], iris.analysis.MEAN, weights=e1_grid_areas
    )

    # Plot the datasets
    qplt.plot(e1_mean, label="E1 scenario", lw=1.5, color="blue")
    qplt.plot(a1b_mean, label="A1B-Image scenario", lw=1.5, color="red")

    # Draw a horizontal line showing the pre-industrial mean
    plt.axhline(
        y=pre_industrial_mean.data,
        color="gray",
        linestyle="dashed",
        label="pre-industrial",
        lw=1.5,
    )

    # Constrain the period 1860-1999 and extract the observed data from a1b
    constraint = iris.Constraint(
        time=lambda cell: 1860 <= cell.point.year <= 1999
    )
    observed = a1b_mean.extract(constraint)

    # Assert that this data set is the same as the e1 scenario:
    # they share data up to the 1999 cut off.
    assert np.all(np.isclose(observed.data, e1_mean.extract(constraint).data))

    # Plot the observed data
    qplt.plot(observed, label="observed", color="black", lw=1.5)

    # Add a legend and title
    plt.legend(loc="upper left")
    plt.title("North American mean air temperature", fontsize=18)

    plt.xlabel("Time / year")
    plt.grid()
    iplt.show()


if __name__ == "__main__":
    main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""Unit tests for :class:`iris.fileformat.ff.ArakawaC`."""

# Import iris.tests first so that some things can be initialised before
# importing anything else.
import iris.tests as tests

import numpy as np

from iris.fileformats._ff import ArakawaC


class Test__x_vectors(tests.IrisTest):
    def _test(self, column, horiz_grid_type, xp, xu):
        reals = np.arange(6) + 100
        grid = ArakawaC(column, None, reals, horiz_grid_type)
        result_xp, result_xu = grid._x_vectors()
        self.assertArrayEqual(result_xp, xp)
        self.assertArrayEqual(result_xu, xu)

    def test_none(self):
        self._test(column=None, horiz_grid_type=None, xp=None, xu=None)

    def test_1d(self):
        self._test(
            column=np.array([[0], [1], [2], [3]]),
            horiz_grid_type=None,
            xp=np.array([0, 1, 2, 3]),
            xu=None,
        )

    def test_2d_no_wrap(self):
        self._test(
            column=np.array([[0, 0], [1, 10], [2, 20], [3, 30]]),
            horiz_grid_type=1,
            xp=np.array([0, 1, 2, 3]),
            xu=np.array([0, 10, 20, 30]),
        )

    def test_2d_with_wrap(self):
        self._test(
            column=np.array([[0, 0], [1, 10], [2, 20], [3, 30]]),
            horiz_grid_type=0,
            xp=np.array([0, 1, 2, 3]),
            xu=np.array([0, 10, 20]),
        )


class Test_regular_x(tests.IrisTest):
    def _test(self, subgrid, bzx, bdx):
        grid = ArakawaC(None, None, [4.0, None, None, -5.0, None, None], None)
        result_bzx, result_bdx = grid.regular_x(subgrid)
        self.assertEqual(result_bzx, bzx)
        self.assertEqual(result_bdx, bdx)

    def test_theta_subgrid(self):
        self._test(1, -9.0, 4.0)

    def test_u_subgrid(self):
        self._test(11, -7.0, 4.0)


class Test_regular_y(tests.IrisTest):
    def _test(self, v_offset, subgrid, bzy, bdy):
        grid = ArakawaC(None, None, [None, 4.0, 45.0, None, None, None], None)
        grid._v_offset = v_offset
        result_bzy, result_bdy = grid.regular_y(subgrid)
        self.assertEqual(result_bzy, bzy)
        self.assertEqual(result_bdy, bdy)

    def test_theta_subgrid_NewDynamics(self):
        self._test(0.5, 1, 41.0, 4.0)

    def test_v_subgrid_NewDynamics(self):
        self._test(0.5, 11, 43.0, 4.0)

    def test_theta_subgrid_ENDGame(self):
        self._test(-0.5, 1, 41.0, 4.0)

    def test_v_subgrid_ENDGame(self):
        self._test(-0.5, 11, 39.0, 4.0)


if __name__ == "__main__":
    tests.main()
